# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.rings.finite_rings
r"""
Evenly distributed sets in finite fields

This module consists of a simple class :class:`EvenlyDistributedSetsBacktracker`. Its main
purpose is to iterate through the evenly distributed sets of a given finite
field.

The naive backtracker implemented here is not directly used to generate
difference family as even for small parameters it already takes time to run.
Instead, its output has been stored into a database
:mod:`sage.combinat.designs.database`. If the backtracker is improved, then one
might want to update this database with more values.

Classes and methods
-------------------
"""

cimport cython

from libc.limits cimport UINT_MAX
from libc.string cimport memset, memcpy
from memory_allocator cimport MemoryAllocator

from sage.rings.integer cimport smallInteger

from sage.categories.fields import Fields


cdef class EvenlyDistributedSetsBacktracker:
    r"""
    Set of evenly distributed subsets in finite fields.

        **Definition:** Let `K` be a finite field of cardinality `q` and `k` an
        integer so that `k(k-1)` divides `q-1`. Let `H = K^*` be the
        multiplicative group of invertible elements in `K`. A `k`-*evenly
        distributed set* in `K` is a set `B = \{b_1, b_2, \ldots, b_k\}` of `k`
        elements of `K` so that the `k(k-1)` differences `\Delta B = \{b_i -
        b_j; i \not= j\}` hit each coset modulo `H^{2(q-1)/(k(k-1))}` exactly
        twice.

    Evenly distributed sets were introduced by Wilson [Wi72]_ (see also
    [BJL99-1]_, Chapter VII.6). He proved that for any `k`, and for any prime power
    `q` large enough such that `k(k-1)` divides `k` there exists a `k`-evenly
    distributed set in the field of cardinality `q`. This existence result based
    on a counting argument (using Dirichlet theorem) does not provide a simple
    method to generate them.

    From a `k`-evenly distributed set, it is straightforward to build a
    `(q,k,1)`-difference family (see :meth:`to_difference_family`). Another
    approach to generate a difference family, somehow dual to this one, is via
    radical difference family (see in particular
    :func:`~sage.combinat.designs.difference_family.radical_difference_family`
    from the module :mod:`~sage.combinat.designs.difference_family`).

    By default, this backtracker only considers evenly distributed sets up to
    affine automorphisms, i.e. `B` is considered equivalent to `s B + t` for any
    invertible element `s` and any element `t` in the field `K`. Note that the
    set of differences is just multiplicatively translated by `s` as `\Delta (s
    B + t) = s (\Delta B)`, and so that `B` is an evenly distributed set if and
    only if `sB` is one too. This behaviour can be modified via the argument
    ``up_to_isomorphism`` (see the input section and the examples below).

    INPUT:

    - ``K`` -- a finite field of cardinality `q`

    - ``k`` -- positive integer such that `k(k-1)` divides `q-1`

    - ``up_to_isomorphism`` -- boolean (default: ``True``); whether only consider
      evenly distributed sets up to automorphisms of the field of the form
      `x \mapsto ax + b`. If set to ``False`` then the iteration is over all
      evenly distributed sets that contain ``0`` and ``1``.

    - ``check`` -- boolean (default: ``False``); whether you want to check
      intermediate steps of the iterator. This is mainly intended for debugging
      purpose. Setting it to ``True`` will considerably slow the iteration.

    EXAMPLES:

    The main part of the code is contained in the iterator. To get one evenly
    distributed set just do::

        sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker
        sage: E = EvenlyDistributedSetsBacktracker(Zmod(151),6)
        sage: B = E.an_element()
        sage: B
        [0, 1, 69, 36, 57, 89]

    The class has a method to convert it to a difference family::

        sage: E.to_difference_family(B)
        [[0, 1, 69, 36, 57, 89],
         [0, 132, 48, 71, 125, 121],
         [0, 59, 145, 10, 41, 117],
         [0, 87, 114, 112, 127, 42],
         [0, 8, 99, 137, 3, 108]]

    It is also possible to run over all evenly distributed sets::

        sage: E = EvenlyDistributedSetsBacktracker(Zmod(13), 4, up_to_isomorphism=False)
        sage: for B in E: print(B)
        [0, 1, 11, 5]
        [0, 1, 4, 6]
        [0, 1, 9, 3]
        [0, 1, 8, 10]

        sage: E = EvenlyDistributedSetsBacktracker(Zmod(13), 4, up_to_isomorphism=True)
        sage: for B in E: print(B)
        [0, 1, 11, 5]


    Or only count them::

        sage: for k in range(13, 200, 12):
        ....:     if is_prime_power(k):
        ....:         K = GF(k,'a')
        ....:         E1 = EvenlyDistributedSetsBacktracker(K, 4, False)
        ....:         E2 = EvenlyDistributedSetsBacktracker(K, 4, True)
        ....:         print("{:3} {:3} {:3}".format(k, E1.cardinality(), E2.cardinality()))
         13   4   1
         25  40   4
         37  12   1
         49  24   2
         61  12   1
         73  48   4
         97  64   6
        109  72   6
        121 240  20
        157  96   8
        169 240  20
        181 204  17
        193 336  28

    Note that by definition, the number of evenly distributed sets up to
    isomorphisms is at most `k(k-1)` times smaller than without isomorphisms.
    But it might not be exactly `k(k-1)` as some of them might have symmetries::

        sage: B = EvenlyDistributedSetsBacktracker(Zmod(13), 4).an_element()
        sage: B
        [0, 1, 11, 5]
        sage: [9*x + 5 for x in B]
        [5, 1, 0, 11]
        sage: [3*x + 11 for x in B]
        [11, 1, 5, 0]
    """
    # PYTHON DATA
    cdef K                       # the underlying field
    cdef list list_K             # the elements of K  (i -> x)
    cdef dict K_to_int           # inverse of list_K  (x -> i)

    # FLAGS
    cdef int count               # do we count or do we iterate
    cdef int check               # do we need to check (debug)
    cdef int up_to_isom          # do we care only about isomorphisms

    # STATIC DATA
    cdef unsigned int q          # cardinality of the field
    cdef unsigned int k          # size of the subsets
    cdef unsigned int e          # k(k-1)/2
    cdef unsigned int m          # (q-1) / e
    cdef unsigned int ** diff    # qxq array: diff[x][y]  = x - y
    cdef unsigned int ** ratio   # qxq array: ratio[x][y] = x / y
    cdef unsigned int * min_orb  # q array  : min_orb[x]  = min {x, 1-x, 1/x,
                                 #                   1/(1-x), (x-1)/x, x/(x-1)}

    # DYNAMIC DATA
    cdef unsigned int * B        # current stack of elements of {0,...,q-1}
    cdef unsigned int * cosets   # e array: cosets of differences of elts in B
    cdef unsigned int * t        # e array: temporary variable for updates

    # MANAGEMENT OF MEMORY
    cdef MemoryAllocator mem

    def __init__(self, K, k, up_to_isomorphism=True, check=False):
        r"""
        TESTS::

            sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker

            sage: EvenlyDistributedSetsBacktracker(Zmod(4),2)
            Traceback (most recent call last):
            ...
            ValueError: Ring of integers modulo 4 is not a field

            sage: EvenlyDistributedSetsBacktracker(Zmod(71),7)
            Traceback (most recent call last):
            ...
            ValueError: k(k-1)=42 does not divide q-1=70

        For `q=421` which is congruent to 1 modulo `12`, `20`, `30` and `42` we
        run backtracker with the ``check`` argument set to ``True``::

            sage: for _ in EvenlyDistributedSetsBacktracker(Zmod(421), 4, check=True):
            ....:     pass
            sage: for _ in EvenlyDistributedSetsBacktracker(Zmod(421), 5, check=True):
            ....:    pass
            sage: for _ in EvenlyDistributedSetsBacktracker(Zmod(421), 6, check=True):
            ....:    pass
            sage: for _ in EvenlyDistributedSetsBacktracker(Zmod(421), 7, check=True):
            ....:    pass
        """
        self.check      = bool(check)
        self.up_to_isom = bool(up_to_isomorphism)
        self.count      = 0

        cdef unsigned int i,j

        if K not in Fields():
            raise ValueError(f"{K} is not a field")
        cdef unsigned int q = K.cardinality()
        cdef unsigned int e = k*(k-1)/2
        if (q-1) % (2*e):
            raise ValueError("k(k-1)={} does not divide q-1={}".format(k*(k-1),q-1))
        cdef unsigned int m = (q - 1) // e

        self.q = q
        self.e = e
        self.k = k
        self.m = (q - 1) // e
        self.K = K

        self.mem = MemoryAllocator()
        self.diff = <unsigned int **> self.mem.calloc(q, sizeof(unsigned int *))
        self.diff[0] = <unsigned int *> self.mem.malloc(q*q*sizeof(unsigned int))
        for i in range(1, self.q):
            self.diff[i] = self.diff[i-1] + q

        self.ratio = <unsigned int **> self.mem.calloc(q, sizeof(unsigned int *))
        self.ratio[0] = <unsigned int *> self.mem.malloc(q*q*sizeof(unsigned int))
        for i in range(1, self.q):
            self.ratio[i] = self.ratio[i-1] + q

        self.B = <unsigned int *> self.mem.malloc(k*sizeof(unsigned int))
        self.min_orb = <unsigned int *> self.mem.malloc(q*sizeof(unsigned int))
        self.cosets = <unsigned int *> self.mem.malloc(e*sizeof(unsigned int))
        self.t = <unsigned int *> self.mem.malloc(e*sizeof(unsigned int))

        x = K.multiplicative_generator()
        list_K = []
        for i in range(e):
            list_K.extend(sorted(x**(j*e+i) for j in range(m)))
        list_K.append(K.zero())
        self.list_K = list_K
        K_to_int = self.K_to_int = {y:i for i,y in enumerate(list_K)}

        zero = K.zero()
        one = K.one()
        assert self.K_to_int[zero] == q-1
        assert self.K_to_int[one] == 0
        assert set(K) == set(list_K)

        self.min_orb[0] = self.min_orb[q-1] = 0
        for i,x in enumerate(self.list_K):
            if x != zero and x != one:
                self.min_orb[i] = min(K_to_int[z] for z in
                        [x, one/x, one-x, one/(one-x), (x-one)/x, x/(x-one)])
            for j,y in enumerate(self.list_K):
                self.diff[i][j]  = K_to_int[x-y]
                if y:
                    self.ratio[i][j] = K_to_int[x/y]
                else:
                    self.ratio[i][j] = UINT_MAX

    def to_difference_family(self, B, check=True):
        r"""
        Given an evenly distributed set ``B`` convert it to a difference family.

        As for any `x\in K^*=H` we have `|\Delta B \cap x
        H^{2(q-1)/(k(k-1))}|=2` (see :class:`EvenlyDistributedSetsBacktracker`),
        the difference family is produced as `\{xB:x\in H^{2(q-1)/(k(k-1))}\}`

        This method is useful if you want to obtain the difference family from
        the output of the iterator.

        INPUT:

        - ``B`` -- an evenly distributed set

        - ``check`` -- boolean (default: ``True``); whether to check the result

        EXAMPLES::

            sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker
            sage: E = EvenlyDistributedSetsBacktracker(Zmod(41),5)
            sage: B = E.an_element(); B
            [0, 1, 13, 38, 31]
            sage: D = E.to_difference_family(B); D
            [[0, 1, 13, 38, 31], [0, 32, 6, 27, 8]]

            sage: from sage.combinat.designs.difference_family import is_difference_family
            sage: is_difference_family(Zmod(41),D,41,5,1)
            True

        Setting ``check`` to ``False`` is much faster::

            sage: timeit("df = E.to_difference_family(B, check=True)") # random
            625 loops, best of 3: 117 µs per loop

            sage: timeit("df = E.to_difference_family(B, check=False)")  # random
            625 loops, best of 3: 1.83 µs per loop
        """
        xe = self.K.multiplicative_generator() ** (self.e)
        df = [[xe**j*b for b in B] for j in range((self.q-1)/(2*self.e))]
        if check:
            from sage.combinat.designs.difference_family import is_difference_family
            if not is_difference_family(self.K, df, self.q, self.k, 1):
                raise RuntimeError("a wrong evenly distributed set was "
                        "produced by the Sage library for the parameters:\n"
                        "  q={}  k={}\n"
                        "Please send an e-mail to "
                        "sage-devel@googlegroups.com".format(self.q, self.k))
        return df

    def an_element(self):
        r"""
        Return an evenly distributed set.

        If there is no such subset raise an
        :class:`~sage.categories.sets_cat.EmptySetError`.

        EXAMPLES::

            sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker

            sage: E = EvenlyDistributedSetsBacktracker(Zmod(41),5)
            sage: E.an_element()
            [0, 1, 13, 38, 31]

            sage: E = EvenlyDistributedSetsBacktracker(Zmod(61),6)
            sage: E.an_element()
            Traceback (most recent call last):
            ...
            EmptySetError: no 6-evenly distributed set in Ring of integers modulo 61
        """
        from sage.categories.sets_cat import EmptySetError
        it = iter(self)
        try:
            B = next(it)
        except StopIteration:
            raise EmptySetError("no {}-evenly distributed set in {}".format(self.k,self.K))
        self.to_difference_family(B, check=True) # check the validity
        return B

    def __repr__(self):
        r"""
        A string representative.

        EXAMPLES::

            sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker

            sage: EvenlyDistributedSetsBacktracker(GF(25,'a'), 4)
            4-evenly distributed sets (up to isomorphism) in Finite Field in a of size 5^2
            sage: EvenlyDistributedSetsBacktracker(GF(25,'a'), 4,
            ....:                                  up_to_isomorphism=False)
            4-evenly distributed sets in Finite Field in a of size 5^2
        """
        return "{}-evenly distributed sets {} in {}".format(
                self.k,
                '(up to isomorphism)' if self.up_to_isom else '',
                self.K)

    def cardinality(self):
        r"""
        Return the number of evenly distributed sets.

        Use with precaution as there can be a lot of such sets and this method
        might be very long to answer!

        EXAMPLES::

            sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker

            sage: E = EvenlyDistributedSetsBacktracker(GF(25,'a'), 4); E
            4-evenly distributed sets (up to isomorphism)
             in Finite Field in a of size 5^2
            sage: E.cardinality()
            4

            sage: E = EvenlyDistributedSetsBacktracker(GF(25,'a'), 4,
            ....:                                      up_to_isomorphism=False)
            sage: E.cardinality()
            40
        """
        cdef n = 0
        self.count = 1
        for a in self:
            n += a
        self.count = 0
        return smallInteger(n)

    def _B_relabelled_copies(self):
        r"""
        Check whether ``self.B`` is minimal among its relabelization.

        If `B=\{b_1,...,b_k\}` is an evenly distributed set and contains `0` and
        `1`, then for any two distinct `i,j` we define `f_{ij} : x \mapsto
        (x-b_j)/(b_i-b_j)` which maps `B` on another evenly distributed set of
        size `k` containing `0` and `1`. For each pair `i,j` we consider check
        whether the set `f_{ij}(B)` is smaller than `B`.

        This is an internal function and should only be call by the backtracker
        implemented in the method ``__iter__``.

        OUTPUT: ``False`` if ``self.B`` is not minimal

        - the list of evenly distributed sets isomorphic to ``self.B``
          given as a list of tuples if ``self.up_to_isom=0`` or list
          containing only ``self.B`` as a tuple if ``self.up_to_isom=1``.

        TESTS::

            sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker
            sage: E = EvenlyDistributedSetsBacktracker(Zmod(13), 4, up_to_isomorphism=True)
            sage: E.cardinality()   # indirect doctest
            1
            sage: E = EvenlyDistributedSetsBacktracker(Zmod(13), 4, up_to_isomorphism=False)
            sage: E.cardinality()   # indirect doctest
            4

        .. NOTE::

            this method is not seriously optimized. The main goal of this backtracker
            is to generate one evenly distributed set. In that case, this method
            will be called only once.
        """
        cdef unsigned int i,j,k,tmp1,tmp2,verify
        cdef list B = [self.B[i] for i in range(1, self.k)]
        B.append(self.q-1)
        cdef list BB = [None]*self.k
        cdef set relabs = set([tuple(B)])

        #  z -> (z - B[i]) / (B[j] - B[i])
        for i in range(self.k):
            for j in range(self.k):
                if i == j:
                    continue
                tmp1 = self.diff[self.B[j]][self.B[i]]

                verify = 0
                for k in range(self.k):
                    if k == i:
                        BB[k] = self.q-1
                    elif k == j:
                        BB[k] = 0
                    else:
                        tmp2 = self.ratio[self.diff[self.B[k]][self.B[i]]][tmp1]
                        if tmp2 == 0 or tmp2 == self.q-1 or tmp2 < self.B[2]:
                            # the backtracker should never build a set which by
                            # relabelling is strictly smaller than B[:3]
                            raise RuntimeError("there is a problem got tmp2={}".format(tmp2,self.B[2]))
                        elif tmp2 == self.B[2]:
                            verify = 1
                        BB[k] = tmp2

                if verify:
                    BB.sort()
                    if BB < B:
                        return False

                if not self.up_to_isom:
                    if not verify:
                        BB.sort()
                    relabs.add(tuple(BB))

        return sorted(relabs)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __iter__(self):
        r"""
        Iterator through all evenly distributed sets that start with `[0,1]`.

        EXAMPLES::

            sage: from sage.combinat.designs.evenly_distributed_sets import EvenlyDistributedSetsBacktracker

            sage: E = EvenlyDistributedSetsBacktracker(Zmod(13),4)
            sage: for B in E:
            ....:     print(B)
            [0, 1, 11, 5]
        """
        cdef unsigned int k_m_1 = self.k - 1
        cdef unsigned int q_m_1 = self.q - 1
        cdef unsigned int m = self.m

        # in the list B we store the candidate for being an e.d.s.
        # we always have B[0] = 0 and B[1] = 1
        # because 0 is in B, the cosets of the elements of B must be
        # disjoint.
        cdef unsigned int kk = 2
        cdef unsigned int * B  = self.B
        B[0] = q_m_1  # the element 0 in K
        B[1] = 0      # the element 1 in K

        memset(self.cosets, 0, self.e * sizeof(unsigned int))

        self.cosets[0] = 1  # coset 0 is hit by the difference 1-0

        cdef unsigned int x = m
        while True:
            if self.check:
                self._check_cosets(kk)
                if x < m or x >= q_m_1:
                    raise RuntimeError("got x < m or x > q_m_1 (x={})".format(x))
                if self.cosets[x/m]:
                    raise RuntimeError("got x={} in an already occupied coset".format(x))

            # try to append x
            B[kk] = x
            if self._check_last_element(kk):
                if kk == k_m_1:
                    ans = self._B_relabelled_copies()

                    if self.check and ans:
                        for a in ans:
                            r = [self.list_K[q_m_1]] + [self.list_K[a[r]] for r in range(k_m_1)]
                            self.to_difference_family(r, check=True)

                    if ans is False:
                        pass
                    elif self.count:
                        yield len(ans)
                    else:
                        for a in ans:
                            yield [self.list_K[q_m_1]] + [self.list_K[a[r]] for r in range(k_m_1)]

                    # remove the differences created by x and increment
                    for j in range(kk):
                        self.cosets[ self.diff[x][B[j]] / m ] = 0
                    x += 1
                else:
                    kk += 1
                    x += m - x%m
            else:
                x += 1

            if self.check:
                self._check_cosets(kk)

            # now we determine the next element x to be tested
            while True:
                if kk == 1:
                    return
                elif x == q_m_1:
                    kk -= 1
                    x = self.B[kk]
                    # remove the differences created by x and increment
                    for j in range(kk):
                        self.cosets[ self.diff[x][B[j]] / m ] = 0
                    x += 1
                    if self.check:
                        self._check_cosets(kk)
                elif self.cosets[x / m]:
                    x += m - x%m
                elif kk == 2:
                    if self.min_orb[x] < x:
                        x += 1
                    else:
                        break
                else:
                    if self.min_orb[x] < B[2]:
                        x += 1
                    else:
                        break

    @cython.cdivision(True)
    cdef inline int _check_last_element(self, unsigned int kk) except -1:
        r"""
        Add the element ``x`` to ``B`` in position kk if the resulting set is
        still evenly distributed.

        OUTPUT:

            1 if the element was added, and 0 otherwise.
        """
        cdef unsigned int i, j, x_m_i, x_m_j
        cdef unsigned int m = self.m
        cdef unsigned int * B = self.B
        cdef unsigned int ** diff = self.diff
        cdef unsigned int x = B[kk]

        # We check two things:
        # 1. that the newly created differences x-B[i] will not be in a coset
        #    already occuppied
        #
        # 2. that by applying some automorphisms we will not get an
        #    element smaller than B[2].
        #
        #    We should test all linear functions that send a subset of the form
        #    {x, B[i], B[j]} to some {0, 1, ?}.
        #
        #    Note that if {x, B[i], B[j]} can be mapped to {0, 1, z} by some
        #    function, then it can also be mapped to all {0, 1, z'} where z'=
        #    1/z, 1-z, 1/(1-z), (z-1)/z and z/(z-1). The attribute
        #    'min_orbit[z]' is exactly the minimum among these values.
        #
        #    So, it is enough to test one of these functions. We choose t -> (x
        #    - t)/ (x - B[j]) (that maps x to 0 and B[j] to 1). Its value at
        #    B[i] is just z = (x - B[i]) / (x - B[j]).
        #
        #    In the special case when kk=2, or equivalently when we are testing if x
        #    fits as a new B[2], then we just check that x is the minimum among
        #    {x, 1/x, 1-x, 1/(1-x), (x-1)/x and x/(x-1)}.

        if self.cosets[diff[x][0] / m] == 1:
            return 0

        self.cosets[x / m] = 1
        for i in range(2,kk):
            x_m_i = diff[x][B[i]]

            # 1. check that the difference x-B[i] was not already in an
            #    occuppied coset
            if self.cosets[x_m_i / m]:
                self.cosets[x / m] = 0
                return 0

            # 2. check relabeling
            for j in range(i):
                x_m_j = diff[x][B[j]]
                if self.min_orb[self.ratio[x_m_i][x_m_j]] < B[2]:
                    self.cosets[x / m] = 0
                    return 0

        # Now check that the x-B[i] belongs to distinct cosets
        memcpy(self.t, self.cosets, self.e*sizeof(unsigned int))
        for i in range(1,kk):
            x_m_i = diff[x][B[i]] / m
            if self.t[x_m_i]:
                self.cosets[x / m] = 0
                return 0
            self.t[x_m_i] = 1
        self.t, self.cosets = self.cosets, self.t
        return 1

    @cython.cdivision(True)
    cdef int _check_cosets(self, unsigned int kk) except -1:
        r"""
        Sanity check (only for debug purposes).
        """
        cdef unsigned int i,j
        cdef unsigned int m = self.m
        cdef unsigned int c

        # count the number of elements in self.cosets
        c = 0
        for i in range(self.e):
            c += self.cosets[i]
        if c != (kk * (kk-1)) / 2:
            raise RuntimeError("the number of elements in cosets is wrong! Got {} instead of {}.".format(c, (kk*(kk-1))/2))

        for i in range(kk):
            for j in range(i):
                if self.cosets[ self.diff[self.B[i]][self.B[j]] / m ] != 1:
                    raise RuntimeError("self.cosets misses the difference B[{}]-B[{}]".format(i,j))

        return 0



# sage_setup: distribution = sagemath-graphs
r"""
Yang-Baxter Graphs
"""
# ****************************************************************************
#       Copyright (C) 2009 Franco Saliola <saliola@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from copy import copy

from sage.graphs.digraph import DiGraph
from sage.structure.sage_object import SageObject
from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.lazy_import import lazy_import
from sage.combinat.permutation import Permutation

lazy_import('sage.combinat.partition', 'Partition')


def YangBaxterGraph(partition=None, root=None, operators=None):
    r"""
    Construct the Yang-Baxter graph from ``root`` by repeated application of
    ``operators``, or the Yang-Baxter graph associated to ``partition``.

    INPUT:

    The user needs to provide either ``partition`` or both ``root`` and
    ``operators``, where

    - ``partition`` -- a partition of a positive integer

    - ``root`` -- the root vertex

    - ``operator`` -- a function that maps vertices `u` to a list of
      tuples of the form `(v, l)` where `v` is a successor of `u` and `l` is
      the label of the edge from `u` to `v`.

    OUTPUT: either:

    - :class:`YangBaxterGraph_partition` -- if partition is defined
    - :class:`YangBaxterGraph_generic` -- if partition is ``None``

    EXAMPLES:

    The Yang-Baxter graph defined by a partition `[p_1,\dots,p_k]` is
    the labelled directed graph with vertex set obtained by
    bubble-sorting `(p_k-1,p_k-2,\dots,0,\dots,p_1-1,p_1-2,\dots,0)`;
    there is an arrow from `u` to `v` labelled by `i` if `v` is
    obtained by swapping the `i`-th and `(i+1)`-th elements of `u`.
    For example, if the partition is `[3,1]`, then we begin with
    `(0,2,1,0)` and generate all tuples obtained from it by swapping
    two adjacent entries if they are increasing::

        sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
        sage: bubbleswaps = [SwapIncreasingOperator(i) for i in range(3)]
        sage: Y = YangBaxterGraph(root=(0,2,1,0), operators=bubbleswaps); Y
        Yang-Baxter graph with root vertex (0, 2, 1, 0)
        sage: Y.vertices(sort=True)
        [(0, 2, 1, 0), (2, 0, 1, 0), (2, 1, 0, 0)]

    The ``partition`` keyword is a shorthand for the above construction::

        sage: Y = YangBaxterGraph(partition=[3,1]); Y                                   # needs sage.combinat
        Yang-Baxter graph of [3, 1], with top vertex (0, 2, 1, 0)
        sage: Y.vertices(sort=True)
        [(0, 2, 1, 0), (2, 0, 1, 0), (2, 1, 0, 0)]

    The permutahedron can be realized as a Yang-Baxter graph::

        sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
        sage: swappers = [SwapIncreasingOperator(i) for i in range(3)]
        sage: Y = YangBaxterGraph(root=(1,2,3,4), operators=swappers); Y
        Yang-Baxter graph with root vertex (1, 2, 3, 4)
        sage: Y.plot()                                                                  # needs sage.plot
        Graphics object consisting of 97 graphics primitives

    The Cayley graph of a finite group can be realized as a Yang-Baxter graph::

        sage: # needs sage.groups
        sage: def left_multiplication_by(g):
        ....:     return lambda h: h*g
        sage: G = CyclicPermutationGroup(4)
        sage: operators = [ left_multiplication_by(gen) for gen in G.gens() ]
        sage: Y = YangBaxterGraph(root=G.identity(), operators=operators); Y
        Yang-Baxter graph with root vertex ()
        sage: Y.plot(edge_labels=False)                                                 # needs sage.plot
        Graphics object consisting of 9 graphics primitives

        sage: # needs sage.groups
        sage: G = SymmetricGroup(4)
        sage: operators = [left_multiplication_by(gen) for gen in G.gens()]
        sage: Y = YangBaxterGraph(root=G.identity(), operators=operators); Y
        Yang-Baxter graph with root vertex ()
        sage: Y.plot(edge_labels=False)                                                 # needs sage.plot
        Graphics object consisting of 96 graphics primitives

    AUTHORS:

    - Franco Saliola (2009-04-23)
    """
    if partition is None:
        return YangBaxterGraph_generic(root=root, operators=operators)
    return YangBaxterGraph_partition(partition=Partition(partition))

# *********** General class for Yang-Baxter Graphs ***********


class YangBaxterGraph_generic(SageObject):
    def __init__(self, root, operators):
        r"""
        A class to model the Yang-Baxter graph defined by ``root`` and
        ``operators``.

        INPUT:

        - ``root`` -- the root vertex of the graph

        - ``operators`` -- list of callables that map vertices to (new)
          vertices

        .. NOTE::

            This is a lazy implementation: the digraph is only computed
            when it is needed.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops); Y
            Yang-Baxter graph with root vertex (1, 0, 2, 1, 0)
            sage: loads(dumps(Y)) == Y
            True

        AUTHORS:

        - Franco Saliola (2009-04-23)
        """
        self._root = root
        self._operators = operators

    def _successors(self, u):
        r"""
        Return an iterator of tuples for the form ``(op(u), op)``, where ``op``
        is one of the operators defining ``self``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: list(Y._successors((1,0,2,1,0)))
            [((1, 2, 0, 1, 0), Swap-if-increasing at position 1)]
        """
        for op in self._operators:
            v = op(u)
            if v != u:
                yield (v, op)

    def __repr__(self) -> str:
        r"""
        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(2)]
            sage: Y = YangBaxterGraph(root=(1,2,3), operators=ops)
            sage: Y.__repr__()
            'Yang-Baxter graph with root vertex (1, 2, 3)'
        """
        return f"Yang-Baxter graph with root vertex {self._root}"

    @lazy_attribute
    def _digraph(self):
        r"""
        Construct the underlying digraph and store the result as an
        attribute.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(2)]
            sage: Y = YangBaxterGraph(root=(1,2,3), operators=ops)
            sage: Y._digraph
            Digraph on 6 vertices
        """
        digraph = DiGraph()
        digraph.add_vertex(self._root)
        queue = [self._root]
        while queue:
            u = queue.pop()
            for v, l in self._successors(u):
                if v not in digraph:
                    queue.append(v)
                digraph.add_edge(u, v, l)
        return digraph

    def __hash__(self) -> int:
        r"""
        TESTS::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(2)]
            sage: Y = YangBaxterGraph(root=(1,2,3), operators=ops)
            sage: H = hash(Y)
        """
        # TODO: this is ugly but unavoidable: the Yang Baxter graphs are being
        # used in containers but are mutable.
        return hash(self._digraph.copy(immutable=True))

    def __eq__(self, other) -> bool:
        r"""
        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y1 = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y2 = YangBaxterGraph(root=(2,0,2,1,0), operators=ops)
            sage: Y3 = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y1.__eq__(Y2)
            False
            sage: Y2.__eq__(Y2)
            True
            sage: Y1.__eq__(Y1)
            True
            sage: Y3.__eq__(Y1)
            True
            sage: Y3.__eq__(Y2)
            False
        """
        return isinstance(other, YangBaxterGraph_generic) and self._digraph == other._digraph

    def __ne__(self, other) -> bool:
        r"""
        Test non-equality.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y1 = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y2 = YangBaxterGraph(root=(2,0,2,1,0), operators=ops)
            sage: Y3 = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y1.__ne__(Y2)
            True
            sage: Y2.__ne__(Y2)
            False
            sage: Y1.__ne__(Y1)
            False
            sage: Y3.__ne__(Y1)
            False
            sage: Y3.__ne__(Y2)
            True
        """
        return not self == other

    def __iter__(self):
        r"""
        Return an iterator of the vertices in ``self``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: sorted(set(Y))
            [(1, 0, 2, 1, 0), (1, 2, 0, 1, 0), (1, 2, 1, 0, 0), (2, 1, 0, 1, 0), (2, 1, 1, 0, 0)]
        """
        return self._digraph.vertex_iterator()

    def __len__(self):
        r"""
        Return the number of vertices in ``self``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y.__len__()
            5
            sage: ops = [SwapIncreasingOperator(i) for i in range(5)]
            sage: Y = YangBaxterGraph(root=(0,1,0,2,1,0), operators=ops)
            sage: Y.__len__()
            16
        """
        return self._digraph.num_verts()

    def __copy__(self):
        r"""
        Return a copy of ``self``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(3)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops); Y
            Yang-Baxter graph with root vertex (1, 0, 2, 1, 0)
            sage: B = copy(Y); B
            Yang-Baxter graph with root vertex (1, 0, 2, 1, 0)
            sage: Y is B
            False
            sage: Y == B
            True
        """
        Y = self.__class__(self._root, self._operators)
        Y._digraph = copy(self._digraph)
        return Y

    def _edges_in_bfs(self):
        r"""
        Return an iterator of the edges of the digraph traversed in a
        breadth-first search of the vertices beginning at ``self.root()``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: list(Y._edges_in_bfs())
            [((1, 0, 2, 1, 0), (1, 2, 0, 1, 0), Swap-if-increasing at position 1), ((1, 2, 0, 1, 0), (1, 2, 1, 0, 0), Swap-if-increasing at position 2), ((1, 2, 0, 1, 0), (2, 1, 0, 1, 0), Swap-if-increasing at position 0), ((2, 1, 0, 1, 0), (2, 1, 1, 0, 0), Swap-if-increasing at position 2)]
        """
        digraph = self._digraph
        seen = {}
        queue = [self._root]
        seen[self._root] = True
        while queue:
            u = queue.pop()
            l = sorted(digraph.neighbor_out_iterator(u))
            for w in l:
                if w not in seen:
                    seen[w] = True
                    queue.append(w)
                    yield (u, w, digraph.edge_label(u, w))

    def root(self):
        r"""
        Return the root vertex of ``self``.

        If ``self`` is the Yang-Baxter graph of the partition
        `[p_1,p_2,\dots,p_k]`, then this is the vertex
        `(p_k-1,p_k-2,\dots,0,\dots,p_1-1,p_1-2,\dots,0)`.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y.root()
            (1, 0, 2, 1, 0)
            sage: Y = YangBaxterGraph(root=(0,1,0,2,1,0), operators=ops)
            sage: Y.root()
            (0, 1, 0, 2, 1, 0)
            sage: Y = YangBaxterGraph(root=(1,0,3,2,1,0), operators=ops)
            sage: Y.root()
            (1, 0, 3, 2, 1, 0)
            sage: Y = YangBaxterGraph(partition=[3,2])                                  # needs sage.combinat
            sage: Y.root()                                                              # needs sage.combinat
            (1, 0, 2, 1, 0)
        """
        return self._root

    def successors(self, v) -> list:
        r"""
        Return the successors of the vertex ``v``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y.successors(Y.root())
            [(1, 2, 0, 1, 0)]
            sage: sorted(Y.successors((1, 2, 0, 1, 0)))
            [(1, 2, 1, 0, 0), (2, 1, 0, 1, 0)]
        """
        return [a for a, _ in self._successors(v)]

    def plot(self, *args, **kwds):
        r"""
        Plot ``self`` as a digraph.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(4)]
            sage: Y = YangBaxterGraph(root=(1,0,2,1,0), operators=ops)
            sage: Y.plot()                                                              # needs sage.plot
            Graphics object consisting of 16 graphics primitives
            sage: Y.plot(edge_labels=False)                                             # needs sage.plot
            Graphics object consisting of 11 graphics primitives
        """
        if "edge_labels" not in kwds:
            kwds["edge_labels"] = True
        if "vertex_labels" not in kwds:
            kwds["vertex_labels"] = True
        return self._digraph.plot(*args, **kwds)

    def vertices(self, sort=False) -> list:
        r"""
        Return the vertices of ``self``.

        INPUT:

        - ``sort`` -- boolean (default: ``False``); whether to sort the vertices

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(3)]
            sage: Y = YangBaxterGraph(root=(0,2,1,0), operators=ops)
            sage: Y.vertices(sort=True)
            [(0, 2, 1, 0), (2, 0, 1, 0), (2, 1, 0, 0)]
        """
        return sorted(self) if sort else list(self)

    def edges(self):
        r"""
        Return the (labelled) edges of ``self``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(3)]
            sage: Y = YangBaxterGraph(root=(0,2,1,0), operators=ops)
            sage: Y.edges()
            [((0, 2, 1, 0), (2, 0, 1, 0), Swap-if-increasing at position 0), ((2, 0, 1, 0), (2, 1, 0, 0), Swap-if-increasing at position 1)]
        """
        return self._digraph.edges(sort=True)

    def vertex_relabelling_dict(self, v, relabel_operator) -> dict:
        r"""
        Return a dictionary pairing vertices ``u`` of ``self`` with
        the object obtained from ``v`` by applying the
        ``relabel_operator`` along a path from the root to ``u``.

        Note that the root is paired with ``v``.

        INPUT:

        - ``v`` -- an object

        - ``relabel_operator`` -- function mapping a vertex and a label to
          the image of the vertex

        OUTPUT: dictionary pairing vertices with the corresponding image of ``v``

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(3)]
            sage: Y = YangBaxterGraph(root=(0,2,1,0), operators=ops)
            sage: def relabel_operator(op, u):
            ....:     i = op.position()
            ....:     return u[:i] + u[i:i+2][::-1] + u[i+2:]
            sage: Y.vertex_relabelling_dict((1,2,3,4), relabel_operator)
            {(0, 2, 1, 0): (1, 2, 3, 4),
             (2, 0, 1, 0): (2, 1, 3, 4),
             (2, 1, 0, 0): (2, 3, 1, 4)}
        """
        relabelling = {self._root: v}
        for u, w, i in self._edges_in_bfs():
            relabelling[w] = relabel_operator(i, relabelling[u])
        return relabelling

    def relabel_vertices(self, v, relabel_operator, inplace=True):
        r"""
        Relabel the vertices ``u`` of ``self`` by the object obtained
        from ``u`` by applying the ``relabel_operator`` to ``v`` along
        a path from ``self.root()`` to ``u``.

        Note that the ``self.root()`` is paired with ``v``.

        INPUT:

        - ``v`` -- tuple, Permutation, etc.

        - ``inplace`` -- if ``True``, modifies ``self``; otherwise returns a
          modified copy of ``self``

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(3)]
            sage: Y = YangBaxterGraph(root=(0,2,1,0), operators=ops)
            sage: def relabel_op(op, u):
            ....:     i = op.position()
            ....:     return u[:i] + u[i:i+2][::-1] + u[i+2:]
            sage: d = Y.relabel_vertices((1,2,3,4), relabel_op, inplace=False); d
            Yang-Baxter graph with root vertex (1, 2, 3, 4)
            sage: Y.vertices(sort=True)
            [(0, 2, 1, 0), (2, 0, 1, 0), (2, 1, 0, 0)]
            sage: e = Y.relabel_vertices((1,2,3,4), relabel_op); e
            sage: Y.vertices(sort=True)
            [(1, 2, 3, 4), (2, 1, 3, 4), (2, 3, 1, 4)]
        """
        relabelling = self.vertex_relabelling_dict(v, relabel_operator)
        Y = self if inplace else copy(self)
        Y._root = relabelling[Y._root]
        Y._digraph.relabel(relabelling, inplace=True)
        if inplace is False:
            return Y

    def relabel_edges(self, edge_dict, inplace=True):
        r"""
        Relabel the edges of ``self``.

        INPUT:

        - ``edge_dict`` -- dictionary keyed by the (unlabelled) edges

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: ops = [SwapIncreasingOperator(i) for i in range(3)]
            sage: Y = YangBaxterGraph(root=(0,2,1,0), operators=ops)
            sage: def relabel_op(op, u):
            ....:     i = op.position()
            ....:     return u[:i] + u[i:i+2][::-1] + u[i+2:]
            sage: Y.edges()
            [((0, 2, 1, 0), (2, 0, 1, 0), Swap-if-increasing at position 0), ((2, 0, 1, 0), (2, 1, 0, 0), Swap-if-increasing at position 1)]
            sage: d = {((0,2,1,0),(2,0,1,0)):17, ((2,0,1,0),(2,1,0,0)):27}
            sage: Y.relabel_edges(d, inplace=False).edges()
            [((0, 2, 1, 0), (2, 0, 1, 0), 17), ((2, 0, 1, 0), (2, 1, 0, 0), 27)]
            sage: Y.edges()
            [((0, 2, 1, 0), (2, 0, 1, 0), Swap-if-increasing at position 0), ((2, 0, 1, 0), (2, 1, 0, 0), Swap-if-increasing at position 1)]
            sage: Y.relabel_edges(d, inplace=True)
            sage: Y.edges()
            [((0, 2, 1, 0), (2, 0, 1, 0), 17), ((2, 0, 1, 0), (2, 1, 0, 0), 27)]
        """
        Y = self if inplace else copy(self)
        digraph = Y._digraph
        for u, v in digraph.edges(sort=False, labels=False):
            digraph.set_edge_label(u, v, edge_dict[u, v])
        if not inplace:
            return Y


# *********** Yang-Baxter Graphs defined by a partition ***********

class YangBaxterGraph_partition(YangBaxterGraph_generic):
    def __init__(self, partition):
        r"""
        A class to model the Yang-Baxter graph of a partition.

        The Yang-Baxter graph defined by a partition `[p_1,\dots,p_k]`
        is the labelled directed graph with vertex set obtained by
        bubble-sorting `(p_k-1,p_k-2,\dots,0,\dots,p_1-1,p_1-2,\dots,0)`;
        there is an arrow from `u` to `v` labelled by `i` if `v` is
        obtained by swapping the `i`-th and `(i+1)`-th elements of `u`.

        .. NOTE::

            This is a lazy implementation: the digraph is only computed
            when it is needed.

        EXAMPLES::

            sage: Y = YangBaxterGraph(partition=[3,2,1]); Y                             # needs sage.combinat
            Yang-Baxter graph of [3, 2, 1], with top vertex (0, 1, 0, 2, 1, 0)
            sage: loads(dumps(Y)) == Y                                                  # needs sage.combinat
            True

        AUTHORS:

        - Franco Saliola (2009-04-23)
        """
        self._partition = partition
        beta = sorted(self._partition, reverse=True)
        root = sum((tuple(range(b)) for b in beta), ())[::-1]
        operators = [SwapIncreasingOperator(i)
                     for i in range(sum(partition) - 1)]
        super().__init__(root, operators)

    def __repr__(self) -> str:
        r"""
        EXAMPLES::

            sage: Y = YangBaxterGraph(partition=[3,2])                                  # needs sage.combinat
            sage: Y.__repr__()                                                          # needs sage.combinat
            'Yang-Baxter graph of [3, 2], with top vertex (1, 0, 2, 1, 0)'
        """
        return f"Yang-Baxter graph of {self._partition}, with top vertex {self._root}"

    def __copy__(self):
        r"""
        Return a copy of ``self``.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: Y = YangBaxterGraph(partition=[3,2]); Y
            Yang-Baxter graph of [3, 2], with top vertex (1, 0, 2, 1, 0)
            sage: B = copy(Y); B
            Yang-Baxter graph of [3, 2], with top vertex (1, 0, 2, 1, 0)
            sage: Y is B
            False
            sage: Y == B
            True
        """
        Y = self.__class__(self._partition)
        Y._digraph = copy(self._digraph)
        return Y

    @lazy_attribute
    def _digraph(self):
        r"""
        Construct the underlying digraph and store the result as an
        attribute.

        EXAMPLES::

            sage: Y = YangBaxterGraph(partition=[2,1])                                  # needs sage.combinat
            sage: Y._digraph                                                            # needs sage.combinat
            Digraph on 2 vertices
            sage: Y.edges()                                                             # needs sage.combinat
            [((0, 1, 0), (1, 0, 0), Swap positions 0 and 1)]
        """
        digraph = super()._digraph
        for u, v, op in digraph.edges():
            digraph.set_edge_label(u, v, SwapOperator(op.position()))
        return digraph

    @lazy_attribute
    def _vertex_ordering(self):
        r"""
        Return a list of the vertices of ``self``, sorted using
        Python's ``sorted`` method.

        EXAMPLES::

            sage: Y = YangBaxterGraph(partition=[3,2])                                  # needs sage.combinat
            sage: Y._vertex_ordering                                                    # needs sage.combinat
            [(1, 0, 2, 1, 0), (1, 2, 0, 1, 0), (1, 2, 1, 0, 0), (2, 1, 0, 1, 0), (2, 1, 1, 0, 0)]
        """
        return self._digraph.vertices(sort=True)

    def __iter__(self):
        r"""
        Iterate over the vertices ``self``.

        .. NOTE::

            The vertices are first sorted using Python's ``sorted`` command.

        EXAMPLES::

            sage: Y = YangBaxterGraph(partition=[3,2])                                  # needs sage.combinat
            sage: list(Y.__iter__())                                                    # needs sage.combinat
            [(1, 0, 2, 1, 0), (1, 2, 0, 1, 0), (1, 2, 1, 0, 0), (2, 1, 0, 1, 0), (2, 1, 1, 0, 0)]
        """
        yield from self._vertex_ordering

    def _swap_operator(self, operator, u):
        r"""
        Return the image of ``u`` under ``operator``.

        INPUT:

        - ``i`` -- positive integer between ``1`` and ``len(u)-1``, inclusive

        - ``u`` -- tuple, list, permutation, etc.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: Y = YangBaxterGraph(partition=[3,1])
            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: ops = [SwapOperator(i) for i in range(3)]
            sage: [Y._swap_operator(op, (1,2,3,4)) for op in ops]
            [(2, 1, 3, 4), (1, 3, 2, 4), (1, 2, 4, 3)]
            sage: [Y._swap_operator(op, [4,3,2,1]) for op in ops]
            [[3, 4, 2, 1], [4, 2, 3, 1], [4, 3, 1, 2]]
            sage: [Y._swap_operator(op, Permutation([1,2,3,4])) for op in ops]
            [[2, 1, 3, 4], [1, 3, 2, 4], [1, 2, 4, 3]]
        """
        return operator(u)

    def vertex_relabelling_dict(self, v) -> dict:
        r"""
        Return a dictionary pairing vertices ``u`` of ``self`` with the object
        obtained from ``v`` by applying transpositions corresponding to the
        edges labels along a path from the root to ``u``.

        Note that the root is paired with ``v``.

        INPUT:

        - ``v`` -- an object

        OUTPUT: dictionary pairing vertices with the corresponding image of ``v``

        EXAMPLES::

            sage: Y = YangBaxterGraph(partition=[3,1])                                  # needs sage.combinat
            sage: Y.vertex_relabelling_dict((1,2,3,4))                                  # needs sage.combinat
            {(0, 2, 1, 0): (1, 2, 3, 4),
             (2, 0, 1, 0): (2, 1, 3, 4),
             (2, 1, 0, 0): (2, 3, 1, 4)}
            sage: Y.vertex_relabelling_dict((4,3,2,1))                                  # needs sage.combinat
            {(0, 2, 1, 0): (4, 3, 2, 1),
             (2, 0, 1, 0): (3, 4, 2, 1),
             (2, 1, 0, 0): (3, 2, 4, 1)}
        """
        return super().vertex_relabelling_dict(v, self._swap_operator)

    def relabel_vertices(self, v, inplace=True):
        r"""
        Relabel the vertices of ``self`` with the object obtained from
        ``v`` by applying the transpositions corresponding to the edge
        labels along some path from the root to the vertex.

        INPUT:

        - ``v`` -- tuple, Permutation, etc.

        - ``inplace`` -- if ``True``, modifies ``self``; otherwise
          returns a modified copy of ``self``

        EXAMPLES::

            sage: # needs sage.combinat
            sage: Y = YangBaxterGraph(partition=[3,1]); Y
            Yang-Baxter graph of [3, 1], with top vertex (0, 2, 1, 0)
            sage: d = Y.relabel_vertices((1,2,3,4), inplace=False); d
            Digraph on 3 vertices
            sage: Y.vertices(sort=True)
            [(0, 2, 1, 0), (2, 0, 1, 0), (2, 1, 0, 0)]
            sage: e = Y.relabel_vertices((1,2,3,4)); e
            sage: Y.vertices(sort=True)
            [(1, 2, 3, 4), (2, 1, 3, 4), (2, 3, 1, 4)]
        """
        relabelling = self.vertex_relabelling_dict(v)
        if inplace:
            Y = self
            Y._root = relabelling[Y._root]
            Y._digraph.relabel(relabelling, inplace=inplace)
            Y._vertex_ordering = Y._digraph.vertices(sort=True)
            return

        Y = copy(self)
        Y._root = relabelling[Y._root]
        return Y._digraph.relabel(relabelling, inplace=inplace)

# ------------- Some Yang-Baxter operators ------------------


class SwapOperator(SageObject):
    def __init__(self, i):
        r"""
        The operator that swaps the items in positions ``i`` and ``i+1``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s3 = SwapOperator(3)
            sage: s3 == loads(dumps(s3))
            True
        """
        self._position = i

    def __hash__(self) -> int:
        r"""
        TESTS::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s = [SwapOperator(i) for i in range(3)]
            sage: [hash(t) for t in s]
            [0, 1, 2]
        """
        return hash(self._position)

    def __eq__(self, other) -> bool:
        r"""
        Compare two swap operators.

        The comparison is done by comparing the positions.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s = [SwapOperator(i) for i in range(3)]
            sage: s[0] == s[0]
            True
            sage: s[1] == s[0]
            False
        """
        if not isinstance(other, SwapOperator):
            return False
        return self._position == other._position

    def __ne__(self, other) -> bool:
        """
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s = [SwapOperator(i) for i in range(3)]
            sage: s[0] != s[0]
            False
            sage: s[1] != s[0]
            True
        """
        return not (self == other)

    def __repr__(self) -> str:
        r"""
        Representation string.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s3 = SwapOperator(3)
            sage: s3.__repr__()
            'Swap positions 3 and 4'
        """
        pos = self._position
        return f"Swap positions {pos} and {pos + 1}"

    def __str__(self) -> str:
        r"""
        A short str representation (used, for example, in labelling edges of
        graphs).

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s3 = SwapOperator(3)
            sage: s3.__str__()
            '3'
        """
        return "%s" % self._position

    def __call__(self, u):
        r"""
        Return the object obtained from swapping the items in positions
        ``i`` and ``i+1`` of ``u``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s3 = SwapOperator(3)
            sage: s3((1,2,3,4,5))
            (1, 2, 3, 5, 4)
            sage: s3([1,2,3,4,5])
            [1, 2, 3, 5, 4]
        """
        i = self._position
        if isinstance(u, Permutation):
            return Permutation(u[:i] + u[i:i + 2][::-1] + u[i + 2:])
        return type(u)(u[:i] + u[i:i + 2][::-1] + u[i + 2:])

    def position(self):
        r"""
        Return ``i`` where ``self`` is the operator that swaps positions ``i``
        and ``i+1``.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapOperator
            sage: s3 = SwapOperator(3)
            sage: s3.position()
            3
        """
        return self._position


class SwapIncreasingOperator(SwapOperator):
    def __repr__(self) -> str:
        r"""
        Representation string.

        EXAMPLES::

            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: s3 = SwapIncreasingOperator(3)
            sage: s3.__repr__()
            'Swap-if-increasing at position 3'
        """
        return "Swap-if-increasing at position %s" % self._position

    def __call__(self, u):
        r"""
        Return a copy of ``u`` with ``u[i-1]`` and ``u[i]`` swapped if
        ``u[i-1] > u[i]``; otherwise returns ``u``.

        INPUT:

        - ``i`` -- positive integer between ``1`` and ``len(u)-1``, inclusive

        - ``u`` -- tuple, list, permutation, etc.

        EXAMPLES::

            sage: Y = YangBaxterGraph(partition=[2,2])                                  # needs sage.combinat
            sage: from sage.combinat.yang_baxter_graph import SwapIncreasingOperator
            sage: operators = [SwapIncreasingOperator(i) for i in range(3)]
            sage: [op((1,2,3,4)) for op in operators]
            [(2, 1, 3, 4), (1, 3, 2, 4), (1, 2, 4, 3)]
            sage: [op([4,3,2,1]) for op in operators]
            [[4, 3, 2, 1], [4, 3, 2, 1], [4, 3, 2, 1]]
            sage: [op(Permutation([1,3,2,4])) for op in operators]
            [[3, 1, 2, 4], [1, 3, 2, 4], [1, 3, 4, 2]]
        """
        i = self._position
        j = i + 1
        if u[i] < u[j]:
            v = list(u)
            v[j], v[i] = v[i], v[j]
            if isinstance(u, Permutation):
                return Permutation(v)
            return type(u)(v)
        return u



# sage_setup: distribution = sagemath-graphs
# *************************************************************************
#        Copyright (C) 2008-9 Robert L. Miller <rlmillster@gmail.com>
#
# Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
#                         https://www.gnu.org/licenses/
# *************************************************************************

from sage.data_structures.bitset cimport bitset_t
from sage.graphs.base.graph_backends cimport GenericGraphBackend
from libc.stdint cimport uint32_t

cdef class CGraph:
    cdef size_t num_verts
    cdef size_t num_arcs
    cdef int *in_degrees
    cdef int *out_degrees
    cdef bitset_t active_vertices

    ###################################
    # Vertex Functions
    ###################################

    cpdef bint has_vertex(self, int n) except -1
    cpdef check_vertex(self, int n)
    cpdef del_vertex(self, int v)
    cpdef int current_allocation(self) noexcept
    cpdef list verts(self)
    cpdef add_vertices(self, verts)
    cdef int del_vertex_unsafe(self, int) except -1
    cpdef realloc(self, int)
    cdef int add_vertex_unsafe(self, int) except -1

    ###################################
    # Edge Functions
    ###################################

    cdef inline int add_arc_unsafe(self, int u, int v) except -1:
        return self.add_arc_label_unsafe(u, v, 0)

    cdef inline int has_arc_unsafe(self, int u, int v) except -1:
        return self.has_arc_label_unsafe(u, v, -1)

    cdef int del_arc_unsafe(self, int, int) except -1

    cpdef add_arc(self, int u, int v)
    cpdef bint has_arc(self, int u, int v) except -1
    cpdef del_all_arcs(self, int u, int v)

    ###################################
    # Labeled Edge Functions
    ###################################

    cdef int add_arc_label_unsafe(self, int, int, int) except -1
    cdef int has_arc_label_unsafe(self, int, int, int) except -1
    cdef int del_arc_label_unsafe(self, int, int, int) except -1
    cdef int arc_label_unsafe(self, int, int) except -1
    cdef int all_arcs_unsafe(self, int, int, int *, int) except -1

    cpdef int arc_label(self, int u, int v) noexcept
    cpdef list all_arcs(self, int u, int v)
    cpdef del_arc_label(self, int u, int v, int l)
    cpdef bint has_arc_label(self, int u, int v, int l) noexcept

    ###################################
    # Neighbor Functions
    ###################################

    cdef int out_neighbors_unsafe(self, int, int *, int) except -2
    cdef int in_neighbors_unsafe(self, int, int *, int) except -2

    cdef inline int _next_neighbor_unsafe(self, int v, int u, bint out, int* l) except -2:
        if out:
            return self.next_out_neighbor_unsafe(v, u, l)
        else:
            return self.next_in_neighbor_unsafe(v, u, l)

    cdef int next_out_neighbor_unsafe(self, int, int, int*) except -2
    cdef int next_in_neighbor_unsafe(self, int, int, int*) except -2
    cdef adjacency_sequence_out(self, int n, int *vertices, int v, int* sequence)
    cdef adjacency_sequence_in(self, int n, int *vertices, int v, int* sequence)
    cpdef list in_neighbors(self, int v)
    cpdef list out_neighbors(self, int u)


cdef class CGraphBackend(GenericGraphBackend):
    cdef int get_vertex(self, u) except ? -2
    cdef int get_vertex_checked(self, u) except ? -2
    cdef vertex_label(self, int u_int)
    cdef int check_labelled_vertex(self, u, bint reverse) except ? -1
    # cdef CGraph _cg  # a child class should declare this accordingly
    cdef bint _directed
    cdef dict vertex_labels
    cdef dict vertex_ints
    cdef dict edge_labels
    cdef bint _loops
    cdef bint _multiple_edges
    cdef CGraph cg(self)
    cpdef add_edge(self, object u, object v, object l, bint directed)
    cpdef del_edge(self, object u, object v, object l, bint directed)
    cdef bint _has_labeled_edge_unsafe(self, int, int, object) except -1
    cdef bint _delete_edge_before_adding(self) noexcept
    cdef int new_edge_label(self, object l) except -1
    cdef int free_edge_label(self, int l_int) except -1
    cdef int _use_edge_iterator_on_subgraph(self, CGraphBackend other, object vertices, const int modus) except -1
    cdef list _all_edge_labels(self, int u, int v, uint32_t* edge=*)



# sage_setup: distribution = sagemath-graphs
#*****************************************************************************
#       Copyright (C) 2006 - 2011 Robert L. Miller <rlmillster@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.groups.perm_gps.partn_ref.data_structures cimport *
from sage.graphs.base.c_graph cimport CGraph
from sage.groups.perm_gps.partn_ref.automorphism_group_canonical_label cimport (
    get_aut_gp_and_can_lab, aut_gp_and_can_lab, agcl_work_space,
    allocate_agcl_output, deallocate_agcl_output,
    allocate_agcl_work_space, deallocate_agcl_work_space)
from sage.groups.perm_gps.partn_ref.canonical_augmentation cimport (iterator,
    canonical_generator_data, allocate_cgd, deallocate_cgd,
    canonical_generator_next,
    setup_canonical_generator, start_canonical_generator)
from sage.groups.perm_gps.partn_ref.refinement_sets cimport (subset, free_subset, all_set_children_are_equivalent,
    refine_set, compare_sets, generate_child_subsets, apply_subset_aug,
    canonical_set_parent, allocate_sgd, deallocate_sgd, allocate_subset_gen, free_subset_gen,
    setup_set_gen, subset_generator_next, subset_generator_data, allocate_subset_gen_2)


cdef class GraphStruct:
    cdef CGraph G
    cdef bint directed
    cdef bint loops
    cdef bint use_indicator
    cdef int *scratch # length 3n+1


cdef struct dg_edge_gen_data:
    iterator *edge_iterator
    void *graph



# sage_setup: distribution = sagemath-graphs
from libcpp.vector cimport vector

from sage.structure.sage_object cimport SageObject
from sage.graphs.base.c_graph cimport CGraph

cdef void extended_lex_BFS(
        CGraph cg, vector[int] &sigma, vector[int] *sigma_inv,
        int initial_v_int, vector[int] *pred, vector[size_t] *xslice_len,
        vector[vector[int]] *lex_label) except *

cdef class SliceDecomposition(SageObject):
    cdef tuple sigma
    cdef dict sigma_inv
    cdef vector[size_t] xslice_len
    cdef dict lex_label
    cdef object _graph_class
    cdef object _underlying_graph



# sage_setup: distribution = sagemath-graphs
from sage.misc.lazy_import import lazy_import

lazy_import('sage.sandpiles.sandpile', 'Sandpile')
lazy_import('sage.sandpiles.sandpile', 'SandpileDivisor')
lazy_import('sage.sandpiles.sandpile', 'SandpileConfig')
lazy_import('sage.sandpiles.sandpile', 'firing_graph')
lazy_import('sage.sandpiles.sandpile', 'parallel_firing_graph')
lazy_import('sage.sandpiles.sandpile', 'wilmes_algorithm')
lazy_import('sage.sandpiles.sandpile', 'triangle_sandpile')

lazy_import('sage.sandpiles.examples', 'sandpiles')
del lazy_import



# sage_setup: distribution = sagemath-graphs
r"""
Common digraphs

All digraphs in Sage can be built through the ``digraphs`` object. In order to
build a circuit on 15 elements, one can do::

    sage: g = digraphs.Circuit(15)

To get a circulant graph on 10 vertices in which a vertex `i` has `i+2` and
`i+3` as outneighbors::

    sage: p = digraphs.Circulant(10,[2,3])

More interestingly, one can get the list of all digraphs that Sage knows how to
build by typing ``digraphs.`` in Sage and then hitting :kbd:`Tab`.

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~DiGraphGenerators.ButterflyGraph`      | Return a `n`-dimensional butterfly graph.
    :meth:`~DiGraphGenerators.Circuit`             | Return the circuit on `n` vertices.
    :meth:`~DiGraphGenerators.Circulant`           | Return a circulant digraph on `n` vertices from a set of integers.
    :meth:`~DiGraphGenerators.Complete`            | Return a complete digraph on `n` vertices.
    :meth:`~DiGraphGenerators.DeBruijn`            | Return the De Bruijn digraph with parameters `k,n`.
    :meth:`~DiGraphGenerators.GeneralizedDeBruijn` | Return the generalized de Bruijn digraph of order `n` and degree `d`.
    :meth:`~DiGraphGenerators.ImaseItoh`           | Return the digraph of Imase and Itoh of order `n` and degree `d`.
    :meth:`~DiGraphGenerators.Kautz`               | Return the Kautz digraph of degree `d` and diameter `D`.
    :meth:`~DiGraphGenerators.nauty_directg`       | Return an iterator yielding digraphs using nauty's ``directg`` program.
    :meth:`~DiGraphGenerators.nauty_posetg`        | Return an iterator yielding Hasse diagrams of posets using nauty's ``genposetg`` program.
    :meth:`~DiGraphGenerators.Paley`               | Return a Paley digraph on `q` vertices.
    :meth:`~DiGraphGenerators.Path`                | Return a directed path on `n` vertices.
    :meth:`~DiGraphGenerators.RandomDirectedAcyclicGraph` | Return a random (weighted) directed acyclic graph of order `n`.
    :meth:`~DiGraphGenerators.RandomDirectedGNC`   | Return a random growing network with copying (GNC) digraph with `n` vertices.
    :meth:`~DiGraphGenerators.RandomDirectedGNM`   | Return a random labelled digraph on `n` nodes and `m` arcs.
    :meth:`~DiGraphGenerators.RandomDirectedGNP`   | Return a random digraph on `n` nodes.
    :meth:`~DiGraphGenerators.RandomDirectedGN`    | Return a random growing network (GN) digraph with `n` vertices.
    :meth:`~DiGraphGenerators.RandomDirectedGNR`   | Return a random growing network with redirection (GNR) digraph.
    :meth:`~DiGraphGenerators.RandomSemiComplete`  | Return a random semi-complete digraph of order `n`.
    :meth:`~DiGraphGenerators.RandomTournament`    | Return a random tournament on `n` vertices.
    :meth:`~DiGraphGenerators.TransitiveTournament`| Return a transitive tournament on `n` vertices.
    :meth:`~DiGraphGenerators.tournaments_nauty`   | Iterator over all tournaments on `n` vertices using Nauty.


AUTHORS:

- Robert L. Miller (2006)
- Emily A. Kirkman (2006)
- Michael C. Yurko (2009)
- David Coudert    (2012)
- Janmenjaya Panda (2024)

Functions and methods
---------------------
"""
# ****************************************************************************
#           Copyright (C) 2006 Robert L. Miller <rlmillster@gmail.com>
#                              and Emily A. Kirkman
#           Copyright (C) 2009 Michael C. Yurko <myurko@gmail.com>
#           Copyright (C) 2012 David Coudert <david.coudert@inria.fr>
#           Copyright (C) 2024 Janmenjaya Panda <janmenjaya.panda.22@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import sys
import subprocess

from sage.cpython.string import bytes_to_str
from sage.misc.randstate import current_randstate
from sage.graphs.digraph import DiGraph
from sage.graphs.graph import Graph


class DiGraphGenerators:
    r"""
    A class consisting of constructors for several common digraphs,
    including orderly generation of isomorphism class representatives.

    A list of all graphs and graph structures in this database is available via
    tab completion. Type ``digraphs.`` and then hit :kbd:`Tab` to see which
    digraphs are available.

    The docstrings include educational information about each named
    digraph with the hopes that this class can be used as a reference.

    The constructors currently in this class include::

                Random Directed Graphs:
                    - RandomDirectedAcyclicGraph
                    - RandomDirectedGN
                    - RandomDirectedGNC
                    - RandomDirectedGNP
                    - RandomDirectedGNM
                    - RandomDirectedGNR
                    - RandomTournament
                    - RandomSemiComplete

                Families of Graphs:
                    - Complete
                    - DeBruijn
                    - GeneralizedDeBruijn
                    - Kautz
                    - Path
                    - ImaseItoh
                    - RandomTournament
                    - TransitiveTournament
                    - tournaments_nauty


    ORDERLY GENERATION: digraphs(vertices, property=lambda x: True,
    augment='edges', size=None)

    Accesses the generator of isomorphism class representatives [McK1998]_.
    Iterates over distinct, exhaustive representatives.

    INPUT:

    - ``vertices`` -- natural number or ``None`` to infinitely generate bigger
      and bigger digraphs

    - ``property`` -- any property to be tested on digraphs before generation

    - ``augment`` -- choices:

      - ``'vertices'`` -- augments by adding a vertex, and edges incident to
        that vertex. In this case, all digraphs on *up to* n=vertices are
        generated. If for any digraph G satisfying the property, every subgraph,
        obtained from G by deleting one vertex and only edges incident to that
        vertex, satisfies the property, then this will generate all digraphs
        with that property. If this does not hold, then all the digraphs
        generated will satisfy the property, but there will be some missing.

      - ``'edges'`` -- augments a fixed number of vertices by adding one
        edge. In this case, all digraphs on *exactly* n=vertices are
        generated. If for any graph G satisfying the property, every subgraph,
        obtained from G by deleting one edge but not the vertices incident to
        that edge, satisfies the property, then this will generate all digraphs
        with that property. If this does not hold, then all the digraphs
        generated will satisfy the property, but there will be some missing.

    - ``implementation`` -- which underlying implementation to use
      (see DiGraph?)

    - ``sparse`` -- boolean (default: ``True``); whether to use a sparse or
      dense data structure. See the documentation of
      :class:`~sage.graphs.graph.Graph`.

    EXAMPLES:

    Print digraphs on 2 or less vertices::

        sage: for D in digraphs(2, augment='vertices'):
        ....:     print(D)
        Digraph on 0 vertices
        Digraph on 1 vertex
        Digraph on 2 vertices
        Digraph on 2 vertices
        Digraph on 2 vertices

    Print digraphs on 3 vertices::

        sage: for D in digraphs(3):
        ....:     print(D)
        Digraph on 3 vertices
        Digraph on 3 vertices
        ...
        Digraph on 3 vertices
        Digraph on 3 vertices

    Generate all digraphs with 4 vertices and 3 edges::

        sage: L = digraphs(4, size=3)
        sage: len(list(L))
        13

    Generate all digraphs with 4 vertices and up to 3 edges::

        sage: L = list(digraphs(4, lambda G: G.size() <= 3))
        sage: len(L)
        20
        sage: graphs_list.show_graphs(L)        # long time                             # needs sage.plot

    Generate all digraphs with degree at most 2, up to 5 vertices::

        sage: property = lambda G: (max([G.degree(v) for v in G] + [0]) <= 2)
        sage: L = list(digraphs(5, property, augment='vertices'))
        sage: len(L)
        75

    Generate digraphs on the fly (see http://oeis.org/classic/A000273)::

        sage: for i in range(5):
        ....:     print(len(list(digraphs(i))))
        1
        1
        3
        16
        218
    """

    def ButterflyGraph(self, n, vertices='strings', immutable=False):
        r"""
        Return a `n`-dimensional butterfly graph.

        The vertices consist of pairs `(v, i)`, where `v` is an `n`-dimensional
        tuple (vector) with binary entries (or a string representation of such)
        and `i` is an integer in `[0..n]`. A directed edge goes from `(v, i)` to
        `(w, i + 1)` if `v` and `w` are identical except for possibly when `v[i]
        \neq w[i]`.

        A butterfly graph has `(2^n)(n+1)` vertices and `n2^{n+1}` edges.

        INPUT:

        - ``n`` -- nonnegative integer; the dimension of the butterfly graph

        - ``vertices`` -- string (default: ``'strings'``); specifies whether the
          vertices are zero-one strings (default) or tuples over GF(2)
          (``vertices='vectors'``)

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        EXAMPLES::

            sage: digraphs.ButterflyGraph(2).edges(sort=True, labels=False)
            [(('00', 0), ('00', 1)),
             (('00', 0), ('10', 1)),
             (('00', 1), ('00', 2)),
             (('00', 1), ('01', 2)),
             (('01', 0), ('01', 1)),
             (('01', 0), ('11', 1)),
             (('01', 1), ('00', 2)),
             (('01', 1), ('01', 2)),
             (('10', 0), ('00', 1)),
             (('10', 0), ('10', 1)),
             (('10', 1), ('10', 2)),
             (('10', 1), ('11', 2)),
             (('11', 0), ('01', 1)),
             (('11', 0), ('11', 1)),
             (('11', 1), ('10', 2)),
             (('11', 1), ('11', 2))]
            sage: digraphs.ButterflyGraph(2, vertices='vectors').edges(sort=True,       # needs sage.modules sage.rings.finite_rings
            ....:                                                      labels=False)
            [(((0, 0), 0), ((0, 0), 1)),
             (((0, 0), 0), ((1, 0), 1)),
             (((0, 0), 1), ((0, 0), 2)),
             (((0, 0), 1), ((0, 1), 2)),
             (((0, 1), 0), ((0, 1), 1)),
             (((0, 1), 0), ((1, 1), 1)),
             (((0, 1), 1), ((0, 0), 2)),
             (((0, 1), 1), ((0, 1), 2)),
             (((1, 0), 0), ((0, 0), 1)),
             (((1, 0), 0), ((1, 0), 1)),
             (((1, 0), 1), ((1, 0), 2)),
             (((1, 0), 1), ((1, 1), 2)),
             (((1, 1), 0), ((0, 1), 1)),
             (((1, 1), 0), ((1, 1), 1)),
             (((1, 1), 1), ((1, 0), 2)),
             (((1, 1), 1), ((1, 1), 2))]
            sage: pos = digraphs.ButterflyGraph(2).get_pos()
            sage: pos['11', 0]
            (0, 0)

        TESTS::

            sage: digraphs.ButterflyGraph(0)
            0-dimensional Butterfly: Digraph on 0 vertices
            sage: digraphs.ButterflyGraph(-1)
            Traceback (most recent call last):
            ...
            ValueError: the number of dimensions must be positive
        """
        if n == 0:
            return DiGraph(name="0-dimensional Butterfly")
        if n < 0:
            raise ValueError("the number of dimensions must be positive")

        # We could switch to Sage integers to handle arbitrary n.
        if vertices == 'strings':
            if n >= 31:
                raise NotImplementedError("vertices='strings' is only valid for n <= 30")
            from sage.graphs.generic_graph_pyx import int_to_binary_string
            V = []
            E = []
            for v in range(2 ** n):
                bv = int_to_binary_string(v)
                # pad and reverse the string
                padded_bv = ('0' * (n - len(bv)) + bv)[::-1]
                V.append(padded_bv)
                for i in range(n):
                    w = v
                    w ^= (1 << i)   # push 1 to the left by i and xor with w
                    bw = int_to_binary_string(w)
                    padded_bw = ('0' * (n - len(bw)) + bw)[::-1]
                    E.append(((padded_bv, i), (padded_bv, i + 1)))
                    E.append(((padded_bv, i), (padded_bw, i + 1)))
        elif vertices == 'vectors':
            from sage.modules.free_module import VectorSpace
            from sage.rings.finite_rings.finite_field_constructor import FiniteField
            from copy import copy
            V = []
            E = []
            for v in VectorSpace(FiniteField(2), n):
                # We must call tuple since vectors are mutable.  To obtain a
                # vector from the tuple tv, just call vector(tv).
                tv = tuple(v)
                V.append(tv)
                w = copy(v)
                for i in range(n):
                    w[i] += 1  # Flip the ith bit
                    E.append(((tv, i), (tv, i + 1)))
                    E.append(((tv, i), (tuple(w), i + 1)))
                    w[i] += 1  # Flip the ith bit back
        else:
            raise NotImplementedError("vertices must be 'strings' or 'vectors'")
        # Set position of vertices
        pos = dict()
        dec = 2**n // n
        for i, v in enumerate(sorted(V, reverse=True)):
            for x in range(n + 1):
                pos[v, x] = (dec * x, i)
        return DiGraph([pos.keys(), E], format='vertices_and_edges', pos=pos,
                       name="{}-dimensional Butterfly".format(n),
                       immutable=immutable)

    def Path(self, n, immutable=False):
        r"""
        Return a directed path on `n` vertices.

        INPUT:

        - ``n`` -- integer; number of vertices in the path

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        EXAMPLES::

            sage: g = digraphs.Path(5)
            sage: g.vertices(sort=True)
            [0, 1, 2, 3, 4]
            sage: g.size()
            4
            sage: g.automorphism_group().cardinality()                                  # needs sage.groups
            1
        """
        g = DiGraph([range(n), zip(range(n - 1), range(1, n))],
                    format='vertices_and_edges', name='Path',
                    immutable=immutable)
        g.set_pos({i: (i, 0) for i in range(n)})
        return g

    def StronglyRegular(self, n, immutable=False):
        r"""
        Return a Strongly Regular digraph with `n` vertices.

        The adjacency matrix of the graph is constructed from a skew Hadamard
        matrix of order `n+1`. These graphs were first constructed in [Duv1988]_.

        INPUT:

        - ``n`` -- integer; the number of vertices of the digraph

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        .. SEEALSO::

            - :func:`sage.combinat.matrices.hadamard_matrix.skew_hadamard_matrix`
            - :meth:`Paley`

        EXAMPLES:

        A Strongly Regular digraph satisfies the condition `AJ = JA = kJ` where
        `A` is the adjacency matrix::

            sage: # needs sage.combinat sage.modules
            sage: g = digraphs.StronglyRegular(7); g
            Strongly regular digraph: Digraph on 7 vertices
            sage: A = g.adjacency_matrix()*ones_matrix(7)
            sage: B = ones_matrix(7)*g.adjacency_matrix()
            sage: A == B == A[0, 0]*ones_matrix(7)
            True

        TESTS:

        Wrong parameter::

            sage: digraphs.StronglyRegular(73)                                          # needs sage.combinat sage.modules
            Traceback (most recent call last):
            ...
            ValueError: strongly regular digraph with 73 vertices not yet implemented
        """
        from sage.combinat.matrices.hadamard_matrix import skew_hadamard_matrix
        from sage.matrix.constructor import ones_matrix, identity_matrix
        if skew_hadamard_matrix(n + 1, existence=True) is not True:
            raise ValueError(f'strongly regular digraph with {n} vertices not yet implemented')

        H = skew_hadamard_matrix(n + 1, skew_normalize=True)
        M = H[1:, 1:]
        M = (M + ones_matrix(n)) / 2 - identity_matrix(n)
        return DiGraph(M, format='adjacency_matrix', immutable=immutable,
                       name='Strongly regular digraph')

    def Paley(self, q, immutable=False):
        r"""
        Return a Paley digraph on `q` vertices.

        Parameter `q` must be the power of a prime number and congruent to 3 mod
        4.

        INPUT:

        - ``q`` -- integer; the number of vertices of the digraph

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        .. SEEALSO::

            - :wikipedia:`Paley_graph`
            - :meth:`~sage.graphs.graph_generators.GraphGenerators.PaleyGraph`

        EXAMPLES:

        A Paley digraph has `n * (n-1) / 2` edges, its underlying graph is a
        clique, and so it is a tournament::

            sage: g = digraphs.Paley(7); g                                              # needs sage.rings.finite_rings
            Paley digraph with parameter 7: Digraph on 7 vertices
            sage: g.size() == g.order() * (g.order() - 1) / 2                           # needs sage.rings.finite_rings
            True
            sage: g.to_undirected().is_clique()                                         # needs sage.rings.finite_rings
            True

        A Paley digraph is always self-complementary::

            sage: g.complement().is_isomorphic(g)                                       # needs sage.rings.finite_rings
            True

        TESTS:

        Wrong parameter::

            sage: digraphs.Paley(6)
            Traceback (most recent call last):
            ...
            ValueError: parameter q must be a prime power
            sage: digraphs.Paley(5)
            Traceback (most recent call last):
            ...
            ValueError: parameter q must be congruent to 3 mod 4
        """
        from sage.rings.finite_rings.integer_mod import mod
        from sage.rings.finite_rings.finite_field_constructor import FiniteField
        from sage.arith.misc import is_prime_power
        if not is_prime_power(q):
            raise ValueError("parameter q must be a prime power")
        if not mod(q, 4) == 3:
            raise ValueError("parameter q must be congruent to 3 mod 4")
        return DiGraph([FiniteField(q, 'a'),
                        lambda i, j: (i != j) and (j - i).is_square()],
                       format='rule', loops=False, immutable=immutable,
                       name="Paley digraph with parameter {}".format(q))

    def TransitiveTournament(self, n, immutable=False):
        r"""
        Return a transitive tournament on `n` vertices.

        In this tournament there is an edge from `i` to `j` if `i<j`.

        See the :wikipedia:`Tournament_(graph_theory)` for more information.

        INPUT:

        - ``n`` -- integer; number of vertices in the tournament

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        EXAMPLES::

            sage: g = digraphs.TransitiveTournament(5)
            sage: g.vertices(sort=True)
            [0, 1, 2, 3, 4]
            sage: g.size()
            10
            sage: g.automorphism_group().cardinality()                                  # needs sage.groups
            1

        .. SEEALSO::

            - :wikipedia:`Tournament_(graph_theory)`
            - :meth:`~sage.graphs.digraph.DiGraph.is_tournament`
            - :meth:`~sage.graphs.digraph.DiGraph.is_transitive`
            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.RandomTournament`

        TESTS::

            sage: digraphs.TransitiveTournament(-1)
            Traceback (most recent call last):
            ...
            ValueError: the number of vertices cannot be strictly negative
        """
        if n < 0:
            raise ValueError('the number of vertices cannot be strictly negative')

        from itertools import combinations
        g = DiGraph([range(n), combinations(range(n), 2)],
                    format='vertices_and_edges', immutable=immutable,
                    name="Transitive Tournament")
        g._circle_embedding(list(range(n)))
        return g

    def RandomTournament(self, n, immutable=False):
        r"""
        Return a random tournament on `n` vertices.

        For every pair of vertices, the tournament has an edge from
        `i` to `j` with probability `1/2`, otherwise it has an edge
        from `j` to `i`.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        EXAMPLES::

            sage: T = digraphs.RandomTournament(10); T
            Random Tournament: Digraph on 10 vertices
            sage: T.size() == binomial(10, 2)                                           # needs sage.symbolic
            True
            sage: T.is_tournament()
            True
            sage: digraphs.RandomTournament(-1)
            Traceback (most recent call last):
            ...
            ValueError: the number of vertices cannot be strictly negative

        .. SEEALSO::

            - :wikipedia:`Tournament_(graph_theory)`
            - :meth:`~sage.graphs.digraph.DiGraph.is_tournament`
            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.TransitiveTournament`
            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.Complete`
            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.RandomSemiComplete`
        """
        if n < 0:
            raise ValueError('the number of vertices cannot be strictly negative')

        from itertools import combinations
        from sage.misc.prandom import getrandbits

        bits = getrandbits(n * (n - 1) // 2)
        edges = ((i, j) if (bits >> k) & 1 else (j, i)
                 for k, (i, j) in enumerate(combinations(range(n), 2)))
        g = DiGraph([range(n), edges], format='vertices_and_edges',
                    immutable=immutable, name="Random Tournament")

        g._circle_embedding(list(range(n)))

        return g

    def tournaments_nauty(self, n,
                          min_out_degree=None, max_out_degree=None,
                          strongly_connected=False, debug=False, options="",
                          immutable=False):
        r"""
        Iterator over all tournaments on `n` vertices using Nauty.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``min_out_degree``, ``max_out_degree`` -- integers; if set to
          ``None`` (default), then the min/max out-degree is not constrained

        - ``debug`` -- boolean (default: ``False``); if ``True`` the first line
          of gentourng's output to standard error is captured and the first call
          to the generator's ``next()`` function will return this line as a
          string.  A line leading with ">A" indicates a successful initiation of
          the program with some information on the arguments, while a line
          beginning with ">E" indicates an error with the input.

        - ``options`` -- string; anything else that should be forwarded as input
          to Nauty's gentourng. See its documentation for more information :
          `<https://pallini.di.uniroma1.it>`_.

        - ``immutable`` -- boolean (default: ``False``); whether to return
          immutable or mutable digraphs.

        EXAMPLES::

            sage: # needs nauty
            sage: for g in digraphs.tournaments_nauty(4):
            ....:     print(g.edges(sort=True, labels = False))
            [(1, 0), (2, 0), (2, 1), (3, 0), (3, 1), (3, 2)]
            [(1, 0), (1, 3), (2, 0), (2, 1), (3, 0), (3, 2)]
            [(0, 2), (1, 0), (2, 1), (3, 0), (3, 1), (3, 2)]
            [(0, 2), (0, 3), (1, 0), (2, 1), (3, 1), (3, 2)]
            sage: tournaments = digraphs.tournaments_nauty
            sage: [len(list(tournaments(x))) for x in range(1,8)]
            [1, 1, 2, 4, 12, 56, 456]
            sage: [len(list(tournaments(x, strongly_connected = True))) for x in range(1,9)]
            [1, 0, 1, 1, 6, 35, 353, 6008]
        """
        nauty_input = options

        if min_out_degree is None:
            min_out_degree = 0
        if max_out_degree is None:
            max_out_degree = n - 1

        nauty_input += " -d" + str(min_out_degree)
        nauty_input += " -D" + str(max_out_degree)

        if strongly_connected:
            nauty_input += " -c"

        nauty_input += " " + str(n) + " "

        import shlex
        from sage.features.nauty import NautyExecutable
        gentourng_path = NautyExecutable("gentourng").absolute_filename()

        sp = subprocess.Popen(shlex.quote(gentourng_path) + " {0}".format(nauty_input),
                              shell=True,
                              stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE, close_fds=True)

        if debug:
            yield sp.stderr.readline()

        def edges(s):
            i = 0
            j = 1
            for b in s[:-1]:
                yield (i, j) if b == '0' else (j, i)

                if j == n - 1:
                    i += 1
                    j = i + 1
                else:
                    j += 1

        gen = sp.stdout
        while True:
            try:
                s = bytes_to_str(next(gen))
            except StopIteration:
                # Exhausted list of graphs from nauty geng
                return

            yield DiGraph([range(n), edges(s)], format='vertices_and_edges',
                          immutable=immutable)

    def nauty_directg(self, graphs, options='', debug=False, immutable=False):
        r"""
        Return an iterator yielding digraphs using nauty's ``directg`` program.

        Description from ``directg --help``:
        Read undirected graphs and orient their edges in all possible ways.
        Edges can be oriented in either or both directions (3 possibilities).
        Isomorphic directed graphs derived from the same input are suppressed.
        If the input graphs are non-isomorphic then the output graphs are also.

        INPUT:

        - ``graphs`` -- a :class:`Graph` or an iterable containing
          :class:`Graph`.  The graph6 string of these graphs is used as an input
          for ``directg``.

        - ``options`` -- string passed to ``directg`` as if it was run at
          a system command line. Available options from ``directg --help``::

            -e<int> | -e<int>:<int>  specify a value or range of the total number of arcs
            -o       orient each edge in only one direction, never both
            -a       only make acyclic orientations (implies -o)
            -f<int>  Use only the subgroup that fixes the first <int> vertices setwise
            -V       only output graphs with nontrivial groups (including exchange of
                     isolated vertices).  The -f option is respected.
            -s<int>/<int>  Make only a fraction of the orientations: The first integer is
                     the part number (first is 0) and the second is the number of
                     parts. Splitting is done per input graph independently.

        - ``debug`` -- boolean (default: ``False``); if ``True`` ``directg``
          standard error and standard output are displayed

        - ``immutable`` -- boolean (default: ``False``); whether to return
          immutable or mutable digraphs.

        EXAMPLES::

            sage: # needs nauty
            sage: gen = graphs.nauty_geng("-c 3")
            sage: dgs = list(digraphs.nauty_directg(gen))
            sage: len(dgs)
            13
            sage: dgs[0]
            Digraph on 3 vertices
            sage: dgs[0]._bit_vector()
            '001001000'
            sage: len(list(digraphs.nauty_directg(graphs.PetersenGraph(), options='-o')))
            324

        Generate non-isomorphic acyclic orientations::

            sage: # needs nauty
            sage: K = graphs.CompleteGraph(4)
            sage: all(d.is_directed_acyclic() for d in digraphs.nauty_directg(K, options='-a'))
            True
            sage: sum(1 for _ in digraphs.nauty_directg(K, options='-a'))
            1
            sage: S = graphs.StarGraph(4)
            sage: all(d.is_directed_acyclic() for d in digraphs.nauty_directg(S, options='-a'))
            True
            sage: sum(1 for _ in digraphs.nauty_directg(S, options='-a'))
            5

        TESTS::

            sage: # needs nauty
            sage: g = digraphs.nauty_directg(graphs.PetersenGraph(), options="-o -G")
            sage: next(g)
            Traceback (most recent call last):
            ...
            ValueError: directg output options [-u|-T|-G] are not allowed
            sage: next(digraphs.nauty_directg(graphs.nauty_geng("-c 3"),
            ....:     options='-o', debug=True))
            &BH?
            &BGO
            &B?o
            &BX?
            &BP_
            <BLANKLINE>
            Digraph on 3 vertices

        .. SEEALSO::

            - :meth:`~sage.graphs.graph.Graph.orientations`
            - :meth:`~sage.graphs.graph.Graph.strong_orientation`
            - :meth:`~sage.graphs.orientations.strong_orientations_iterator`
            - :meth:`~sage.graphs.orientations.random_orientation`
        """
        if '-u' in options or '-T' in options or '-G' in options:
            raise ValueError("directg output options [-u|-T|-G] are not allowed")

        if isinstance(graphs, Graph):
            graphs = [graphs]
        elif not graphs:
            return

        if '-q' not in options:
            options += ' -q'

        # Build directg input (graphs6 format)
        input = ''.join(g.graph6_string() + '\n' for g in graphs)

        import shlex
        from sage.features.nauty import NautyExecutable
        directg_path = NautyExecutable("directg").absolute_filename()

        sub = subprocess.Popen(shlex.quote(directg_path) + ' {0}'.format(options),
                               shell=True,
                               stdout=subprocess.PIPE,
                               stdin=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               encoding='latin-1')
        out, err = sub.communicate(input=input)

        if debug:
            if err:
                print(err)

            if out:
                print(out)

        for line in out.split('\n'):
            # directg return graphs in the digraph6 format.
            # digraph6 is very similar with the dig6 format used in sage :
            # digraph6_string = '&' +  dig6_string
            # digraph6 specifications:
            # http://users.cecs.anu.edu.au/~bdm/data/formats.txt
            if line and line[0] == '&':
                yield DiGraph(line[1:], format='dig6', immutable=immutable)

    def nauty_posetg(self, options='', debug=False, immutable=False):
        r"""
        Return a generator which creates all posets using ``nauty``.

        Here a poset is seen through its Hasse diagram, which is
        an acyclic and transitively reduced digraph.

        INPUT:

        - ``options`` -- string (default: ``""``); a string passed to
          ``genposetg`` as if it was run at a system command line.
          At a minimum, you *must* pass the number of vertices you desire
          and a choice between ``o`` and ``t`` for the output order.

        - ``debug`` -- boolean (default: ``False``); if ``True`` the first line
          of ``genposetg``'s output to standard error is captured and the first
          call to the generator's ``next()`` function will return this line as a
          string. A line leading with ">A" indicates a successful initiation of
          the program with some information on the arguments, while a line
          beginning with ">E" indicates an error with the input.

        - ``immutable`` -- boolean (default: ``False``); whether to return
          immutable or mutable posets.

        The possible options, obtained as output of ``genposetg --help``::

            n: the number of vertices, between 0 and 16
            o: digraph6 output in arbitrary order
            t: digraph6 output in topological order

        EXAMPLES::

            sage: gen = digraphs.nauty_posetg("5 o")                                    # needs nauty
            sage: len(list(gen))                                                        # needs nauty
            63

        This coincides with :oeis:`A000112`.
        """
        import shlex
        from sage.features.nauty import NautyExecutable
        geng_path = NautyExecutable("genposetg").absolute_filename()
        sp = subprocess.Popen(shlex.quote(geng_path) + f" {options}", shell=True,
                              stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE, close_fds=True,
                              encoding='latin-1')
        msg = sp.stderr.readline()
        if debug:
            yield msg
        elif msg.startswith('>E'):
            raise ValueError('wrong format of parameter option')
        gen = sp.stdout
        while True:
            try:
                s = next(gen)
            except StopIteration:
                # Exhausted list of graphs from nauty genposetg
                return
            yield DiGraph(s[1:-1], format='dig6', immutable=immutable)

    def Complete(self, n, loops=False, immutable=False):
        r"""
        Return the complete digraph on `n` vertices.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``loops`` -- boolean (default: ``False``); whether to add loops or
          not, i.e., edges from `u` to itself

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        .. SEEALSO::

            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.RandomSemiComplete`

            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.RandomTournament`

        EXAMPLES::

            sage: n = 10
            sage: G = digraphs.Complete(n); G
            Complete digraph: Digraph on 10 vertices
            sage: G.size() == n*(n-1)
            True
            sage: G = digraphs.Complete(n, loops=True); G
            Complete digraph with loops: Looped digraph on 10 vertices
            sage: G.size() == n*n
            True
            sage: digraphs.Complete(-1)
            Traceback (most recent call last):
            ...
            ValueError: the number of vertices cannot be strictly negative
        """
        if n < 0:
            raise ValueError('the number of vertices cannot be strictly negative')

        edges = ((u, v) for u in range(n) for v in range(n) if u != v or loops)
        G = DiGraph([range(n), edges], format='vertices_and_edges',
                    loops=loops, immutable=immutable,
                    name="Complete digraph" + (" with loops" if loops else ''))

        G._circle_embedding(list(range(n)))

        return G

    def Circuit(self, n, immutable=False):
        r"""
        Return the circuit on `n` vertices.

        The circuit is an oriented
        :meth:`~sage.graphs.graph_generators.GraphGenerators.CycleGraph`.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        EXAMPLES:

        A circuit is the smallest strongly connected digraph::

            sage: circuit = digraphs.Circuit(15)
            sage: len(circuit.strongly_connected_components()) == 1
            True
        """
        if n < 0:
            raise ValueError('the number of vertices cannot be strictly negative')
        if n == 1:
            return DiGraph([(0, 0)], format='list_of_edges', loops=True,
                           immutable=immutable, name='Circuit')

        from itertools import chain
        edges = zip(range(n), chain(range(1, n), [0]))
        g = DiGraph([range(n), edges], format='vertices_and_edges',
                    immutable=immutable, name='Circuit')
        g._circle_embedding(list(range(n)))
        return g

    def Circulant(self, n, integers, immutable=False):
        r"""
        Return a circulant digraph on `n` vertices from a set of integers.

        A circulant digraph of order `n` has an arc from vertex `i` to
        vertex `i+j \pmod{n}`, for each `j` in ``integers``.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``integers`` -- iterable container (list, set, etc.) of integers such
          that there is an edge from `i` to `j` if and only if `(j-i) \pmod{n}`
          is an integer

        - ``immutable`` -- boolean (default: ``False``); whether to return
          an immutable or mutable digraph.

        EXAMPLES:

        Construct and show the circulant graph [3, 5, 7], a digraph on 13
        vertices::

            sage: g = digraphs.Circulant(13, [3, 5, 7])
            sage: g.show()                          # long time                             # needs sage.plot

        The Koh-Tindell digraph [LM2024]_ is the circulant digraph of order 7
        with parameters `[1, 5]`.  This `2`-diregular digraph is
        vertex-transitive but not arc-transitive. The associated bipartite
        digraph of the Koh-Tindell digraph is a Pfaffian orientation of the
        Heawood graph. Construct and show the Koh-Tindell digraph::

            sage: kohTindellDigraph = digraphs.Circulant(7, [1, 5])
            sage: kohTindellDigraph.show()          # long time                             # needs sage.plot

        TESTS:

            sage: digraphs.Circulant(13, [3, 5, 7, "hey"])
            Traceback (most recent call last):
            ...
            ValueError: the list must contain only integers
            sage: digraphs.Circulant(3, [3, 5, 7, 3.4])
            Traceback (most recent call last):
            ...
            ValueError: the list must contain only integers
        """
        from sage.rings.integer_ring import ZZ

        # Bad input and loops
        loops = False
        for i in integers:
            if i not in ZZ:
                raise ValueError("the list must contain only integers")
            if not i % n:
                loops = True

        edges = ((v, (v + j) % n) for j in integers for v in range(n))
        G = DiGraph([range(n), edges], format='vertices_and_edges',
                    loops=loops, immutable=immutable,
                    name="Circulant graph (" + str(integers) + ")")
        G._circle_embedding(list(range(n)))
        return G

    def DeBruijn(self, k, n, vertices='strings'):
        r"""
        Return the De Bruijn digraph with parameters `k,n`.

        The De Bruijn digraph with parameters `k,n` is built upon a set of
        vertices equal to the set of words of length `n` from a dictionary of
        `k` letters.

        In this digraph, there is an arc `w_1w_2` if `w_2` can be obtained from
        `w_1` by removing the leftmost letter and adding a new letter at its
        right end.  For more information, see the :wikipedia:`De_Bruijn_graph`.

        INPUT:

        - ``k`` -- two possibilities for this parameter :
            - An integer equal to the cardinality of the alphabet to use, that
              is, the degree of the digraph to be produced.
            - An iterable object to be used as the set of letters. The degree
              of the resulting digraph is the cardinality of the set of letters.

        - ``n`` -- integer; length of words in the De Bruijn digraph when
          ``vertices == 'strings'``, and also the diameter of the digraph

        - ``vertices`` -- string (default: ``'strings'``); whether the vertices
          are words over an alphabet (default) or integers
          (``vertices='string'``)

        EXAMPLES:

        de Bruijn digraph of degree 2 and diameter 2::

            sage: db = digraphs.DeBruijn(2, 2); db                                      # needs sage.combinat
            De Bruijn digraph (k=2, n=2): Looped digraph on 4 vertices
            sage: db.order(), db.size()                                                 # needs sage.combinat
            (4, 8)
            sage: db.diameter()                                                         # needs sage.combinat
            2

        Building a de Bruijn digraph on a different alphabet::

            sage: # needs sage.combinat
            sage: g = digraphs.DeBruijn(['a', 'b'], 2)
            sage: g.vertices(sort=True)
            ['aa', 'ab', 'ba', 'bb']
            sage: g.is_isomorphic(db)
            True
            sage: g = digraphs.DeBruijn(['AA', 'BB'], 2)
            sage: g.vertices(sort=True)
            ['AA,AA', 'AA,BB', 'BB,AA', 'BB,BB']
            sage: g.is_isomorphic(db)
            True

        TESTS:

        Alphabet of null size or words of length zero::

            sage: digraphs.DeBruijn(5, 0)                                               # needs sage.combinat
            De Bruijn digraph (k=5, n=0): Looped multi-digraph on 1 vertex
            sage: digraphs.DeBruijn(0, 0)                                               # needs sage.combinat
            De Bruijn digraph (k=0, n=0): Looped multi-digraph on 0 vertices

        :issue:`22355`::

            sage: db = digraphs.DeBruijn(2, 2, vertices='strings')                      # needs sage.combinat
            sage: db.vertices(sort=True)                                                # needs sage.combinat
            ['00', '01', '10', '11']
            sage: h = digraphs.DeBruijn(2, 2, vertices='integers')
            sage: h.vertices(sort=True)
            [0, 1, 2, 3]
            sage: db.is_isomorphic(h)                                                   # needs sage.combinat
            True
            sage: digraphs.DeBruijn(0, 0, vertices='integers')
            De Bruijn digraph (k=0, n=0): Looped multi-digraph on 0 vertices
            sage: digraphs.DeBruijn(2, 2, vertices='circles')
            Traceback (most recent call last):
            ...
            ValueError: unknown type for vertices
        """
        from sage.rings.integer import Integer

        if vertices == 'strings':
            from sage.combinat.words.words import Words

            W = Words(list(range(k)) if isinstance(k, Integer) else k, n)
            A = Words(list(range(k)) if isinstance(k, Integer) else k, 1)
            g = DiGraph(loops=True)

            if not n:
                g.allow_multiple_edges(True)
                v = W[0]
                vs = v.string_rep()
                for a in A:
                    g.add_edge(vs, vs, a.string_rep())
            else:
                for w in W:
                    ww = w[1:]
                    ws = w.string_rep()
                    for a in A:
                        g.add_edge(ws, (ww * a).string_rep(), a.string_rep())

        elif vertices == 'integers':
            d = k if isinstance(k, Integer) else len(list(k))
            if not d:
                g = DiGraph(loops=True, multiedges=True)
            else:
                g = digraphs.GeneralizedDeBruijn(d ** n, d)

        else:
            raise ValueError('unknown type for vertices')

        g.name("De Bruijn digraph (k={}, n={})".format(k, n))
        return g

    def GeneralizedDeBruijn(self, n, d):
        r"""
        Return the generalized de Bruijn digraph of order `n` and degree `d`.

        The generalized de Bruijn digraph was defined in [RPK1980]_ [RPK1983]_.
        It has vertex set `V=\{0, 1,..., n-1\}` and there is an arc from vertex
        `u \in V` to all vertices `v \in V` such that `v \equiv (u*d + a)
        \mod{n}` with `0 \leq a < d`.

        When `n = d^{D}`, the generalized de Bruijn digraph is isomorphic to
        the de Bruijn digraph of degree `d` and diameter `D`.

        INPUT:

        - ``n`` -- integer; number of vertices of the digraph (must be at least
          one)

        - ``d`` -- integer; degree of the digraph (must be at least one)

        .. SEEALSO::

            * :meth:`sage.graphs.generic_graph.GenericGraph.is_circulant` --
              checks whether a (di)graph is circulant, and/or returns all
              possible sets of parameters.

        EXAMPLES::

            sage: GB = digraphs.GeneralizedDeBruijn(8, 2)
            sage: GB.is_isomorphic(digraphs.DeBruijn(2, 3), certificate=True)           # needs sage.combinat
            (True, {0: '000', 1: '001', 2: '010', 3: '011',
                    4: '100', 5: '101', 6: '110', 7: '111'})

        TESTS:

        An exception is raised when the degree is less than one::

            sage: G = digraphs.GeneralizedDeBruijn(2, 0)
            Traceback (most recent call last):
            ...
            ValueError: degree must be greater than or equal to one

        An exception is raised when the order of the graph is less than one::

            sage: G = digraphs.GeneralizedDeBruijn(0, 2)
            Traceback (most recent call last):
            ...
            ValueError: order must be greater than or equal to one
        """
        if n < 1:
            raise ValueError("order must be greater than or equal to one")
        if d < 1:
            raise ValueError("degree must be greater than or equal to one")

        GB = DiGraph(n, loops=True, multiedges=True,
                     name="Generalized de Bruijn digraph (n={}, d={})".format(n, d))
        for u in range(n):
            for a in range(u * d, u * d + d):
                GB.add_edge(u, a % n)
        return GB

    def ImaseItoh(self, n, d):
        r"""
        Return the Imase-Itoh digraph of order `n` and degree `d`.

        The Imase-Itoh digraph was defined in [II1983]_. It has vertex set
        `V=\{0, 1,..., n-1\}` and there is an arc from vertex `u \in V` to all
        vertices `v \in V` such that `v \equiv (-u*d-a-1) \mod{n}` with `0 \leq
        a < d`.

        When `n = d^{D}`, the Imase-Itoh digraph is isomorphic to the de Bruijn
        digraph of degree `d` and diameter `D`. When `n = d^{D-1}(d+1)`, the
        Imase-Itoh digraph is isomorphic to the Kautz digraph [Kau1968]_ of
        degree `d` and diameter `D`.

        INPUT:

        - ``n`` -- integer; number of vertices of the digraph (must be greater
          than or equal to two)

        - ``d`` -- integer; degree of the digraph (must be greater than or
          equal to one)

        EXAMPLES::

            sage: II = digraphs.ImaseItoh(8, 2)
            sage: II.is_isomorphic(digraphs.DeBruijn(2, 3), certificate=True)           # needs sage.combinat
            (True, {0: '010', 1: '011', 2: '000', 3: '001',
                    4: '110', 5: '111', 6: '100', 7: '101'})

            sage: II = digraphs.ImaseItoh(12, 2)
            sage: b,D = II.is_isomorphic(digraphs.Kautz(2, 3), certificate=True)        # needs sage.combinat
            sage: b                                                                     # needs sage.combinat
            True
            sage: D   # random isomorphism                                              # needs sage.combinat
            {0: '202', 1: '201', 2: '210', 3: '212', 4: '121',
             5: '120', 6: '102', 7: '101', 8: '010', 9: '012',
             10: '021', 11: '020'}

        TESTS:

        An exception is raised when the degree is less than one::

            sage: G = digraphs.ImaseItoh(2, 0)
            Traceback (most recent call last):
            ...
            ValueError: degree must be greater than or equal to one

        An exception is raised when the order of the graph is less than two::

            sage: G = digraphs.ImaseItoh(1, 2)
            Traceback (most recent call last):
            ...
            ValueError: order must be greater than or equal to two
        """
        if n < 2:
            raise ValueError("order must be greater than or equal to two")
        if d < 1:
            raise ValueError("degree must be greater than or equal to one")

        II = DiGraph(n, loops=True, multiedges=True,
                     name="Imase and Itoh digraph (n={}, d={})".format(n, d))
        for u in range(n):
            for a in range(-u * d - d, -u * d):
                II.add_edge(u, a % n)
        return II

    def Kautz(self, k, D, vertices='strings'):
        r"""
        Return the Kautz digraph of degree `d` and diameter `D`.

        The Kautz digraph has been defined in [Kau1968]_. The Kautz digraph of
        degree `d` and diameter `D` has `d^{D-1}(d+1)` vertices. This digraph
        is built from a set of vertices equal to the set of words of length `D`
        over an alphabet of `d+1` letters such that consecutive letters are
        different. There is an arc from vertex `u` to vertex `v` if `v` can be
        obtained from `u` by removing the leftmost letter and adding a new
        letter, distinct from the rightmost letter of `u`, at the right end.

        The Kautz digraph of degree `d` and diameter `D` is isomorphic to the
        Imase-Itoh digraph [II1983]_ of degree `d` and order `d^{D-1}(d+1)`.

        See the :wikipedia:`Kautz_graph` for more information.

        INPUT:

        - ``k`` -- two possibilities for this parameter. In either case the
          degree must be at least one:

            - An integer equal to the degree of the digraph to be produced,
              that is, the cardinality of the alphabet to be used minus one.
            - An iterable object to be used as the set of letters. The degree
              of the resulting digraph is the cardinality of the set of letters
              minus one.

        - ``D`` -- integer; diameter of the digraph, and length of a vertex
          label when ``vertices == 'strings'`` (must be at least one)

        - ``vertices`` -- string (default: ``'strings'``); whether the vertices
          are words over an alphabet (default) or integers
          (``vertices='strings'``)

        EXAMPLES::

            sage: # needs sage.combinat
            sage: K = digraphs.Kautz(2, 3)
            sage: b, D = K.is_isomorphic(digraphs.ImaseItoh(12, 2), certificate=True)
            sage: b
            True
            sage: D  # random isomorphism
            {'010': 8, '012': 9, '020': 11, '021': 10, '101': 7,  '102': 6,
             '120': 5, '121': 4, '201': 1, '202': 0, '210': 2, '212': 3}

            sage: K = digraphs.Kautz([1,'a','B'], 2)                                    # needs sage.combinat
            sage: K.edges(sort=True)                                                    # needs sage.combinat
            [('1B', 'B1', '1'), ('1B', 'Ba', 'a'), ('1a', 'a1', '1'),
             ('1a', 'aB', 'B'), ('B1', '1B', 'B'), ('B1', '1a', 'a'),
             ('Ba', 'a1', '1'), ('Ba', 'aB', 'B'), ('a1', '1B', 'B'),
             ('a1', '1a', 'a'), ('aB', 'B1', '1'), ('aB', 'Ba', 'a')]

            sage: K = digraphs.Kautz([1,'aA','BB'], 2)                                  # needs sage.combinat
            sage: K.edges(sort=True)                                                    # needs sage.combinat
            [('1,BB', 'BB,1', '1'), ('1,BB', 'BB,aA', 'aA'),
             ('1,aA', 'aA,1', '1'), ('1,aA', 'aA,BB', 'BB'),
             ('BB,1', '1,BB', 'BB'), ('BB,1', '1,aA', 'aA'),
             ('BB,aA', 'aA,1', '1'), ('BB,aA', 'aA,BB', 'BB'),
             ('aA,1', '1,BB', 'BB'), ('aA,1', '1,aA', 'aA'),
             ('aA,BB', 'BB,1', '1'), ('aA,BB', 'BB,aA', 'aA')]

        TESTS:

        An exception is raised when the degree is less than one::

            sage: G = digraphs.Kautz(0, 2)                                              # needs sage.combinat
            Traceback (most recent call last):
            ...
            ValueError: degree must be greater than or equal to one

            sage: G = digraphs.Kautz(['a'], 2)                                          # needs sage.combinat
            Traceback (most recent call last):
            ...
            ValueError: degree must be greater than or equal to one

        An exception is raised when the diameter of the graph is less than
        one::

            sage: G = digraphs.Kautz(2, 0)                                              # needs sage.combinat
            Traceback (most recent call last):
            ...
            ValueError: diameter must be greater than or equal to one

        :issue:`22355`::

            sage: K = digraphs.Kautz(2, 2, vertices='strings')                          # needs sage.combinat
            sage: K.vertices(sort=True)                                                 # needs sage.combinat
            ['01', '02', '10', '12', '20', '21']
            sage: h = digraphs.Kautz(2, 2, vertices='integers')
            sage: h.vertices(sort=True)
            [0, 1, 2, 3, 4, 5]
            sage: h.is_isomorphic(K)                                                    # needs sage.combinat
            True
            sage: h = digraphs.Kautz([1,'aA','BB'], 2, vertices='integers')
            sage: h.is_isomorphic(K)                                                    # needs sage.combinat
            True
            sage: h.vertices(sort=True)
            [0, 1, 2, 3, 4, 5]
            sage: digraphs.Kautz(2, 2, vertices='circles')
            Traceback (most recent call last):
            ...
            ValueError: unknown type for vertices
        """
        if D < 1:
            raise ValueError("diameter must be greater than or equal to one")

        from sage.rings.integer import Integer

        if vertices == 'strings':
            from sage.combinat.words.words import Words

            my_alphabet = Words([str(i) for i in range(k + 1)] if isinstance(k,
                                Integer) else k, 1)
            if my_alphabet.alphabet().cardinality() < 2:
                raise ValueError("degree must be greater than or equal to one")

            # We start building the set of vertices
            V = list(my_alphabet)
            for i in range(D - 1):
                VV = []
                for w in V:
                    VV += [w * a for a in my_alphabet if not w.has_suffix(a)]
                V = VV

            # We now build the set of arcs
            G = DiGraph()
            for u in V:
                us = u.string_rep()
                for a in my_alphabet:
                    if not u.has_suffix(a):
                        G.add_edge(us, (u[1:] * a).string_rep(),
                                   a.string_rep())

        elif vertices == 'integers':
            d = k if isinstance(k, Integer) else (len(list(k)) - 1)
            if d < 1:
                raise ValueError("degree must be greater than or equal to one")
            G = digraphs.ImaseItoh((d + 1) * (d ** (D - 1)), d)

        else:
            raise ValueError('unknown type for vertices')

        G.name("Kautz digraph (k={}, D={})".format(k, D))
        return G

    def RandomDirectedAcyclicGraph(self, n, p, weight_max=None):
        r"""
        Return a random (weighted) directed acyclic graph of order `n`.

        The method starts with the sink vertex and adds vertices one at a time.
        A vertex is connected only to previously defined vertices, and the
        probability of each possible connection is given by the probability `p`.
        The weight of an edge is a random integer between ``1`` and
        ``weight_max``.

        INPUT:

        - ``n`` -- number of nodes of the graph

        - ``p`` -- probability of an edge

        - ``weight_max`` -- (default: ``None``) by default, the returned DAG is
          unweighted. When ``weight_max`` is set to a positive integer, edges
          are assigned a random integer weight between ``1`` and ``weight_max``.

        EXAMPLES::

            sage: D = digraphs.RandomDirectedAcyclicGraph(5, .5); D
            RandomDAG(5, 0.500000000000000): Digraph on 5 vertices
            sage: D.is_directed_acyclic()
            True
            sage: D = digraphs.RandomDirectedAcyclicGraph(5, .5, weight_max=3); D
            RandomWeightedDAG(5, 0.500000000000000, 3): Digraph on 5 vertices
            sage: D.is_directed_acyclic()
            True

        TESTS:

        Check special cases::

            sage: digraphs.RandomDirectedAcyclicGraph(0, .5).order() == 0
            True
            sage: digraphs.RandomDirectedAcyclicGraph(4, 0).size() == 0
            True
            sage: digraphs.RandomDirectedAcyclicGraph(4, 1).size() == 6
            True

        Check that bad inputs are rejected::

            sage: digraphs.RandomDirectedAcyclicGraph(-1, .5)
            Traceback (most recent call last):
            ...
            ValueError: the number of nodes must be positive or null
            sage: digraphs.RandomDirectedAcyclicGraph(5, 1.1)
            Traceback (most recent call last):
            ...
            ValueError: the probability p must be in [0..1]
            sage: digraphs.RandomDirectedAcyclicGraph(5, .5, weight_max=-1)
            Traceback (most recent call last):
            ...
            ValueError: parameter weight_max must be a positive integer
        """
        if n < 0:
            raise ValueError("the number of nodes must be positive or null")
        if 0.0 > p or 1.0 < p:
            raise ValueError("the probability p must be in [0..1]")

        # according the sage.misc.randstate.pyx documentation, random
        # integers are on 31 bits. We thus set the pivot value to p*2^31
        from sage.misc.prandom import randint
        from sage.misc.randstate import random
        RAND_MAX_f = float(1 << 31)
        pp = int(round(float(p * RAND_MAX_f)))

        if weight_max is None:
            D = DiGraph(n, name=f"RandomDAG({n}, {p})")
            D.add_edges((i, j) for i in range(n) for j in range(i) if random() < pp)

        else:
            from sage.rings.integer_ring import ZZ
            if weight_max in ZZ and weight_max < 1:
                raise ValueError("parameter weight_max must be a positive integer")

            D = DiGraph(n, name=f"RandomWeightedDAG({n}, {p}, {weight_max})")
            D.add_edges((i, j, randint(1, weight_max))
                        for i in range(n) for j in range(i) if random() < pp)

        return D

    def RandomDirectedGN(self, n, kernel=None, seed=None):
        r"""
        Return a random growing network (GN) digraph with `n` vertices.

        The digraph is constructed by adding vertices with a link to one
        previously added vertex. The vertex to link to is chosen with a
        preferential attachment model, i.e. probability is proportional to
        degree. The default attachment kernel is a linear function of
        degree. The digraph is always a tree, so in particular it is a
        directed acyclic graph. See [KR2001b]_ for more details.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``kernel`` -- the attachment kernel (default: identity function)

        - ``seed`` -- a ``random.Random`` seed or a Python ``int`` for the
          random number generator (default: ``None``)

        EXAMPLES::

            sage: # needs networkx
            sage: D = digraphs.RandomDirectedGN(25)
            sage: D.num_verts()
            25
            sage: D.num_edges()
            24
            sage: D.is_connected()
            True
            sage: D.parent() is DiGraph
            True
            sage: D.show()                      # long time
        """
        if kernel is None:
            kernel = lambda x: x
        if seed is None:
            seed = int(current_randstate().long_seed() % sys.maxsize)
        import networkx
        return DiGraph(networkx.gn_graph(n, kernel, seed=seed))

    def RandomDirectedGNC(self, n, seed=None):
        r"""
        Return a random growing network with copying (GNC) digraph with `n`
        vertices.

        The digraph is constructed by adding vertices with a link to one
        previously added vertex. The vertex to link to is chosen with a
        preferential attachment model, i.e. probability is proportional to
        degree. The new vertex is also linked to all of the previously
        added vertex's successors. See [KR2005]_ for more details.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``seed`` -- a ``random.Random`` seed or a Python ``int`` for the
          random number generator (default: ``None``)

        EXAMPLES::

            sage: # needs networkx
            sage: D = digraphs.RandomDirectedGNC(25)
            sage: D.is_directed_acyclic()
            True
            sage: D.topological_sort()
            [24, 23, ..., 1, 0]
            sage: D.show()                      # long time
        """
        if seed is None:
            seed = int(current_randstate().long_seed() % sys.maxsize)
        import networkx
        return DiGraph(networkx.gnc_graph(n, seed=seed))

    def RandomDirectedGNP(self, n, p, loops=False, seed=None):
        r"""
        Return a random digraph on `n` nodes.

        Each edge is inserted independently with probability `p`.
        See [ER1959]_ and [Gil1959]_ for more details.

        INPUT:

        - ``n`` -- integer; number of nodes of the digraph

        - ``p`` -- float; probability of an edge

        - ``loops`` -- boolean (default: ``False``); whether the random digraph
          may have loops

        - ``seed`` -- integer (default: ``None``); seed for random number
          generator

        PLOTTING: When plotting, this graph will use the default spring-layout
        algorithm, unless a position dictionary is specified.

        EXAMPLES::

            sage: D = digraphs.RandomDirectedGNP(10, .2)
            sage: D.num_verts()
            10
            sage: D.parent() is DiGraph
            True
        """
        from sage.graphs.graph_generators_pyx import RandomGNP
        if 0.0 > p or 1.0 < p:
            raise ValueError("the probability p must be in [0..1]")

        if seed is None:
            seed = current_randstate().long_seed()

        return RandomGNP(n, p, directed=True, loops=loops, seed=seed)

    def RandomDirectedGNM(self, n, m, loops=False):
        r"""
        Return a random labelled digraph on `n` nodes and `m` arcs.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``m`` -- integer; number of edges

        - ``loops`` -- boolean (default: ``False``); whether to allow loops

        PLOTTING: When plotting, this graph will use the default spring-layout
        algorithm, unless a position dictionary is specified.

        EXAMPLES::

            sage: D = digraphs.RandomDirectedGNM(10, 5)
            sage: D.num_verts()
            10
            sage: D.num_edges()
            5

        With loops::

            sage: D = digraphs.RandomDirectedGNM(10, 100, loops = True)
            sage: D.num_verts()
            10
            sage: D.loops()
            [(0, 0, None), (1, 1, None), (2, 2, None), (3, 3, None),
             (4, 4, None), (5, 5, None), (6, 6, None), (7, 7, None),
             (8, 8, None), (9, 9, None)]

        TESTS::

            sage: digraphs.RandomDirectedGNM(10,-3)
            Traceback (most recent call last):
            ...
            ValueError: the number of edges must satisfy 0 <= m <= n(n-1) when no loops are allowed, and 0 <= m <= n^2 otherwise

            sage: digraphs.RandomDirectedGNM(10,100)
            Traceback (most recent call last):
            ...
            ValueError: the number of edges must satisfy 0 <= m <= n(n-1) when no loops are allowed, and 0 <= m <= n^2 otherwise
        """
        n, m = int(n), int(m)

        # The random graph is built by drawing randomly and uniformly two
        # integers u,v, and adding the corresponding edge if it does not exist,
        # as many times as necessary.

        # When the graph is dense, we actually compute its complement. This will
        # prevent us from drawing the same pair u,v too many times.

        from sage.misc.prandom import _pyrand
        rand = _pyrand()
        D = DiGraph(n, loops=loops)

        # Ensuring the parameters n,m make sense.
        #
        # If the graph is dense, we actually want to build its complement. We
        # update m accordingly.

        good_input = True
        is_dense = False

        if m < 0:
            good_input = False

        if loops:
            if m > n * n:
                good_input = False
            elif 2 * m > n * n:
                is_dense = True
                m = n * n - m

        else:
            if m > n * (n - 1):
                good_input = False
            elif m > (n * (n - 1)) // 2:
                is_dense = True
                m = n * (n - 1) - m

        if not good_input:
            raise ValueError("the number of edges must satisfy 0 <= m <= n(n-1) "
                             "when no loops are allowed, and 0 <= m <= n^2 otherwise")

        # When the given number of edges defines a density larger than 1/2, it
        # should be faster to compute the complement of the graph (less edges to
        # generate), then to return its complement. This being said, the
        # .complement() method for sparse graphs is very slow at the moment.

        # Similarly, it is faster to test whether a pair belongs to a dictionary
        # than to test the adjacency of two vertices in a graph. For these
        # reasons, the following code mainly works on dictionaries.

        adj = {i: dict() for i in range(n)}

        # We fill the dictionary structure, but add the corresponding edge in
        # the graph only if is_dense is False. If it is true, we will add the
        # edges in a second phase.

        while m > 0:

            # It is better to obtain random numbers this way than by calling the
            # randint or randrange method. This, because they are very expensive
            # when trying to compute MANY random integers, and because the
            # following lines is precisely what they do anyway, after checking
            # their parameters are correct.

            u = int(rand.random() * n)
            v = int(rand.random() * n)

            if (u != v or loops) and (v not in adj[u]):
                adj[u][v] = 1
                m -= 1
                if not is_dense:
                    D.add_edge(u, v)

        # If is_dense is True, it means the graph has not been built. We fill D
        # with the complement of the edges stored in the adj dictionary

        if is_dense:
            for u in range(n):
                for v in range(n):
                    if ((u != v) or loops) and (v not in adj[u]):
                        D.add_edge(u, v)

        return D

    def RandomDirectedGNR(self, n, p, seed=None):
        r"""
        Return a random growing network with redirection (GNR) digraph
        with `n` vertices and redirection probability `p`.

        The digraph is constructed by adding vertices with a link to one
        previously added vertex. The vertex to link to is chosen uniformly.
        With probability p, the arc is instead redirected to the successor
        vertex. The digraph is always a tree.
        See [KR2001b]_ for more details.

        INPUT:

        - ``n`` -- integer; number of vertices

        - ``p`` -- redirection probability

        - ``seed`` -- a ``random.Random`` seed or a Python ``int`` for the
          random number generator (default: ``None``)

        EXAMPLES::

            sage: # needs networkx
            sage: D = digraphs.RandomDirectedGNR(25, .2)
            sage: D.is_directed_acyclic()
            True
            sage: D.to_undirected().is_tree()
            True
            sage: D.show()                      # long time                             # needs sage.plot
        """
        if seed is None:
            seed = int(current_randstate().long_seed() % sys.maxsize)
        import networkx
        return DiGraph(networkx.gnr_graph(n, p, seed=seed))

    def RandomSemiComplete(self, n):
        r"""
        Return a random semi-complete digraph on `n` vertices.

        A directed graph `G=(V,E)` is *semi-complete* if for any pair of
        vertices `u` and `v`, there is *at least* one arc between them.

        To generate randomly a semi-complete digraph, we have to ensure, for any
        pair of distinct vertices `u` and `v`, that with probability `1/3` we
        have only arc `uv`, with probability `1/3` we have only arc `vu`, and
        with probability `1/3` we have both arc `uv` and arc `vu`. We do so by
        selecting a random integer `coin` in `[1,3]`. When `coin==1` we select
        only arc `uv`, when `coin==3` we select only arc `vu`, and when
        `coin==2` we select both arcs. In other words, we select arc `uv` when
        `coin\leq 2` and arc `vu` when `coin\geq 2`.

        INPUT:

        - ``n`` -- integer; the number of nodes

        .. SEEALSO::

            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.Complete`

            - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.RandomTournament`

        EXAMPLES::

            sage: SC = digraphs.RandomSemiComplete(10); SC
            Random Semi-Complete digraph: Digraph on 10 vertices
            sage: SC.size() >= binomial(10, 2)                                          # needs sage.symbolic
            True
            sage: digraphs.RandomSemiComplete(-1)
            Traceback (most recent call last):
            ...
            ValueError: the number of vertices cannot be strictly negative
        """
        G = DiGraph(n, name="Random Semi-Complete digraph")

        # For each pair u,v we choose a random number ``coin`` in [1,3].
        # We select edge `(u,v)` if `coin==1` or `coin==2`.
        # We select edge `(v,u)` if `coin==2` or `coin==3`.
        import itertools
        from sage.misc.prandom import randint
        for u, v in itertools.combinations(range(n), 2):
            coin = randint(1, 3)
            if coin <= 2:
                G.add_edge(u, v)
            if coin >= 2:
                G.add_edge(v, u)

        G._circle_embedding(list(range(n)))

        return G

# ##############################################################################
#   DiGraph Iterators
# ##############################################################################

    def __call__(self, vertices=None, property=lambda x: True, augment='edges',
                 size=None, sparse=True, copy=True):
        """
        Access the generator of isomorphism class representatives [McK1998]_.
        Iterates over distinct, exhaustive representatives.

        INPUT:

        - ``vertices`` -- natural number or ``None`` to generate all digraphs

        - ``property`` -- any property to be tested on digraphs before
          generation

        - ``augment`` -- choices:

          - ``'vertices'`` -- augments by adding a vertex, and edges incident to
            that vertex. In this case, all digraphs on up to n=vertices are
            generated. If for any digraph G satisfying the property, every
            subgraph, obtained from G by deleting one vertex and only edges
            incident to that vertex, satisfies the property, then this will
            generate all digraphs with that property. If this does not hold,
            then all the digraphs generated will satisfy the property, but there
            will be some missing.

          - ``'edges'`` -- augments a fixed number of vertices by adding one
            edge. In this case, all digraphs on exactly n=vertices are
            generated. If for any graph G satisfying the property, every
            subgraph, obtained from G by deleting one edge but not the vertices
            incident to that edge, satisfies the property, then this will
            generate all digraphs with that property. If this does not hold,
            then all the digraphs generated will satisfy the property, but there
            will be some missing.

        - ``sparse`` -- boolean (default: ``True``); whether to use a sparse or
          dense data structure. See the documentation of
          :class:`~sage.graphs.graph.Graph`.

        - ``copy`` -- boolean (default: ``True``); whether to make copies of the
          digraphs before returning them. If set to ``False`` the method returns
          the digraph it is working on. The second alternative is faster, but
          modifying any of the digraph instances returned by the method may
          break the function's behaviour, as it is using these digraphs to
          compute the next ones: only use ``copy = False`` when you stick to
          *reading* the digraphs returned.

        EXAMPLES:

        Print digraphs on 2 or less vertices::

            sage: for D in digraphs(2, augment='vertices'):
            ....:     print(D)
            Digraph on 0 vertices
            Digraph on 1 vertex
            Digraph on 2 vertices
            Digraph on 2 vertices
            Digraph on 2 vertices

        Print digraphs on 3 vertices::

            sage: for D in digraphs(3):
            ....:     print(D)
            Digraph on 3 vertices
            Digraph on 3 vertices
            ...
            Digraph on 3 vertices
            Digraph on 3 vertices

        For more examples, see the class level documentation, or type ::

            sage: digraphs?  # not tested
        """
        from copy import copy as copyfun
        if size is not None:
            def extra_property(x):
                return x.size() == size
        else:
            def extra_property(x):
                return True
        if augment == 'vertices':
            if vertices is None:
                raise NotImplementedError

            from sage.graphs.graph_generators import canaug_traverse_vert
            g = DiGraph(sparse=sparse)
            for gg in canaug_traverse_vert(g, [], vertices, property, dig=True, sparse=sparse):
                if extra_property(gg):
                    yield copyfun(gg) if copy else gg

        elif augment == 'edges':

            if vertices is None:
                vertices = 0
                while True:
                    yield from self(vertices, sparse=sparse, copy=copy)
                    vertices += 1

            from sage.graphs.graph_generators import canaug_traverse_edge
            g = DiGraph(vertices, sparse=sparse)
            gens = []
            for i in range(vertices - 1):
                gen = list(range(i))
                gen.append(i + 1)
                gen.append(i)
                gen += list(range(i + 2, vertices))
                gens.append(gen)
            for gg in canaug_traverse_edge(g, gens, property, dig=True, sparse=sparse):
                if extra_property(gg):
                    yield copyfun(gg) if copy else gg
        else:
            raise NotImplementedError()


# Easy access to the graph generators from the command line:
digraphs = DiGraphGenerators()



# sage_setup: distribution = sagemath-graphs
from libc.stdint cimport uint32_t

cdef unsigned short * c_shortest_path_all_pairs(G, vertex_list=*) except NULL
cdef unsigned short * c_distances_all_pairs(G, vertex_list=*) noexcept
cdef all_pairs_shortest_path_BFS(gg,
                                 unsigned short * predecessors,
                                 unsigned short * distances,
                                 uint32_t       * eccentricity,
                                 vertex_list=*)

cdef uint32_t * c_eccentricity(G, vertex_list=*) except NULL
